# ***************************************************************************
# *   Copyright (c) 2023 Syres                                              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw FillTemplateFields Task Dialog."""

__title__ = "TechDrawTools.CommandFillTemplateFields"
__author__ = "Syres"
__url__ = "https://www.freecad.org"
__version__ = "00.02"
__date__ = "2023/12/07"

from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide import QtCore, QtGui
from PySide.QtGui import QLineEdit, QCheckBox
import FreeCAD as App
import FreeCADGui as Gui
import datetime
from datetime import date
import csv
import codecs
from fractions import Fraction
import os.path
import TechDraw

CreatedByChkLst = []
ScaleChkLst = []
LabelChkLst = []
CommentChkLst = []
CompanyChkLst = []
LicenseChkLst = []
CreatedDateChkLst = []
LastModifiedDateChkLst = []
listofkeys = [
    "CreatedByChkLst",
    "ScaleChkLst",
    "LabelChkLst",
    "CommentChkLst",
    "CompanyChkLst",
    "LicenseChkLst",
    "CreatedDateChkLst",
    "LastModifiedDateChkLst",
]

"""Run the following code when the command is activated (button press)."""
file_path = App.getResourceDir() + "Mod/TechDraw/CSVdata/FillTemplateFields.csv"

if os.path.exists(file_path):
    with codecs.open(file_path, encoding="utf-8") as fp:
        reader = csv.DictReader(fp)
        if listofkeys == reader.fieldnames:
            for row in reader:
                CreatedByChkLst.append(row["CreatedByChkLst"])
                ScaleChkLst.append(row["ScaleChkLst"])
                LabelChkLst.append(row["LabelChkLst"])
                CommentChkLst.append(row["CommentChkLst"])
                CompanyChkLst.append(row["CompanyChkLst"])
                LicenseChkLst.append(row["LicenseChkLst"])
                CreatedDateChkLst.append(row["CreatedDateChkLst"])
                LastModifiedDateChkLst.append(row["LastModifiedDateChkLst"])
        else:
            errorMsg = QtCore.QT_TRANSLATE_NOOP(
                "Techdraw_FillTemplateFields",
                " file does not contain the correct field names therefore exiting",
            )
            App.Console.PrintError(file_path + errorMsg + "\n")
else:
    errorMsg = QtCore.QT_TRANSLATE_NOOP(
        "Techdraw_FillTemplateFields",
        " file has not been found therefore exiting",
    )
    App.Console.PrintError(file_path + errorMsg + "\n")

keyLst = []


class TaskFillTemplateFields:
    def __init__(self):
        objs = App.ActiveDocument.findObjects(Type="TechDraw::DrawPage")

        for obj in objs:
            if (
                os.path.exists(file_path)
                and listofkeys == reader.fieldnames
            ):
                self.page = obj
                if obj.Views == []:
                    msgBox = QtGui.QMessageBox()
                    msgTitle = QtCore.QT_TRANSLATE_NOOP(
                        "Techdraw_FillTemplateFields",
                        "View or projection group missing",
                    )
                    msg = QtCore.QT_TRANSLATE_NOOP(
                        "Techdraw_FillTemplateFields",
                        "There must be a view or projection group to"
                        " establish data for the scale field in " + self.page.Label,
                    )
                    msgBox.setText(msg)
                    msgBox.setWindowTitle(msgTitle)
                    msgBox.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
                    msgBox.exec_()
                    break

                projgrp_view = None
                for pageObj in obj.Views:
                    if hasattr(pageObj, "Scale"):
                        # use the scale from the first DVP or DPG encountered to fill the template's
                        # Scale editable text.
                        projgrp_view = pageObj
                        break

                self.texts = self.page.Template.EditableTexts

                self.dialog = None
                self.s1 = None

                self.dialog = QtGui.QDialog()
                self.dialog.resize(1050, 400)
                self.dialog.setWindowTitle(
                    QtCore.QT_TRANSLATE_NOOP(
                        "TechDraw_FillTemplateFields", "Fill Template Fields In "
                    )
                    + self.page.Label
                )
                self.dialog.move(400, 200)
                self.la = QtGui.QGridLayout(self.dialog)
                updateCb = QtCore.QT_TRANSLATE_NOOP(
                    "TechDraw_FillTemplateFields", "Update"
                )
                icon_name = "button_right"
                svg = ":/icons/" + icon_name
                pix = QtGui.QPixmap(svg)
                self.checkBoxList = []
                self.lineTextList = []
                dialogRow = 0
                longestText = 0
                for key, value in self.texts.items():
                    App.Console.PrintLog("{0} = {1} | ".format(key, value))
                    if str(key).lower() in CreatedByChkLst:
                        t1 = QtGui.QLabel(value)
                        self.la.addWidget(t1, dialogRow, 0)
                        self.cb1 = QtGui.QCheckBox(updateCb)
                        self.cb1.setObjectName(key)
                        if hasattr(self.cb1, "setCheckState"):
                            self.cb1.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb1.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb1, dialogRow, 1)
                        u1 = QtGui.QLabel("")
                        u1.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u1, dialogRow, 2)
                        self.s1 = QtGui.QLineEdit()
                        self.s1.setText(App.ActiveDocument.CreatedBy)
                        self.la.addWidget(self.s1, dialogRow, 3)
                        self.cb1.setObjectName(key)
                        keyLst.append(self.cb1.objectName())
                        self.checkBoxList.append(self.cb1)
                        self.lineTextList.append(self.s1)
                        self.cb1.clicked.connect(self.on_cb1_clicked)
                        longestText = max(
                            longestText, len(App.ActiveDocument.CreatedBy)
                        )
                        dialogRow += 1
                    if str(key).lower() in ScaleChkLst and projgrp_view:
                        t2 = QtGui.QLabel(value)
                        self.la.addWidget(t2, dialogRow, 0)
                        self.cb2 = QtGui.QCheckBox(updateCb)
                        self.cb2.setObjectName(key)
                        if hasattr(self.cb2, "setCheckState"):
                            self.cb2.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb2.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb2, dialogRow, 1)
                        u2 = QtGui.QLabel("")
                        u2.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u2, dialogRow, 2)
                        self.s2 = QtGui.QLineEdit()
                        self.la.addWidget(self.s2, dialogRow, 3)
                        self.cb2.setObjectName(key)
                        keyLst.append(self.cb2.objectName())
                        self.checkBoxList.append(self.cb2)
                        self.lineTextList.append(self.s2)
                        self.cb2.clicked.connect(self.on_cb2_clicked)
                        fracScale = TechDraw.nearestFraction(projgrp_view.Scale)
                        self.s2.setText(
                                str(fracScale[0])
                                + " : "
                                + str(fracScale[1]))
                        dialogRow += 1
                    if str(key).lower() in LabelChkLst:
                        t3 = QtGui.QLabel(value)
                        self.la.addWidget(t3, dialogRow, 0)
                        self.cb3 = QtGui.QCheckBox(updateCb)
                        self.cb3.setObjectName(key)
                        if hasattr(self.cb3, "setCheckState"):
                            self.cb3.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb3.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb3, dialogRow, 1)
                        u3 = QtGui.QLabel("")
                        u3.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u3, dialogRow, 2)
                        self.s3 = QtGui.QLineEdit()
                        self.la.addWidget(self.s3, dialogRow, 3)
                        self.s3.setText(App.ActiveDocument.Label)
                        self.cb3.setObjectName(key)
                        keyLst.append(self.cb3.objectName())
                        self.checkBoxList.append(self.cb3)
                        self.lineTextList.append(self.s3)
                        self.cb3.clicked.connect(self.on_cb3_clicked)
                        longestText = max(longestText, len(App.ActiveDocument.Label))
                        dialogRow += 1
                    if str(key).lower() in CommentChkLst:
                        t4 = QtGui.QLabel(value)
                        self.la.addWidget(t4, dialogRow, 0)
                        self.cb4 = QtGui.QCheckBox(updateCb)
                        self.cb4.setObjectName(key)
                        if hasattr(self.cb4, "setCheckState"):
                            self.cb4.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb4.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb4, dialogRow, 1)
                        u4 = QtGui.QLabel("")
                        u4.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u4, dialogRow, 2)
                        self.s4 = QtGui.QLineEdit()
                        self.la.addWidget(self.s4, dialogRow, 3)
                        self.s4.setText(App.ActiveDocument.Comment)
                        self.cb4.setObjectName(key)
                        keyLst.append(self.cb4.objectName())
                        self.checkBoxList.append(self.cb4)
                        self.lineTextList.append(self.s4)
                        self.cb4.clicked.connect(self.on_cb4_clicked)
                        longestText = max(longestText, len(App.ActiveDocument.Comment))
                        dialogRow += 1
                    if str(key).lower() in CompanyChkLst:
                        t5 = QtGui.QLabel(value)
                        self.la.addWidget(t5, dialogRow, 0)
                        self.cb5 = QtGui.QCheckBox(updateCb)
                        self.cb5.setObjectName(key)
                        if hasattr(self.cb5, "setCheckState"):
                            self.cb5.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb5.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb5, dialogRow, 1)
                        u5 = QtGui.QLabel("")
                        u5.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u5, dialogRow, 2)
                        self.s5 = QtGui.QLineEdit()
                        self.la.addWidget(self.s5, dialogRow, 3)
                        self.s5.setText(App.ActiveDocument.Company)
                        self.cb5.setObjectName(key)
                        keyLst.append(self.cb5.objectName())
                        self.checkBoxList.append(self.cb5)
                        self.lineTextList.append(self.s5)
                        self.cb5.clicked.connect(self.on_cb5_clicked)
                        longestText = max(longestText, len(App.ActiveDocument.Company))
                        dialogRow += 1
                    if str(key).lower() in LicenseChkLst:
                        t6 = QtGui.QLabel(value)
                        self.la.addWidget(t6, dialogRow, 0)
                        self.cb6 = QtGui.QCheckBox(updateCb)
                        self.cb6.setObjectName(key)
                        if hasattr(self.cb6, "setCheckState"):
                            self.cb6.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb6.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb6, dialogRow, 1)
                        u6 = QtGui.QLabel("")
                        u6.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u6, dialogRow, 2)
                        self.s6 = QtGui.QLineEdit()
                        self.la.addWidget(self.s6, dialogRow, 3)
                        self.s6.setText(App.ActiveDocument.License)
                        self.cb6.setObjectName(key)
                        keyLst.append(self.cb6.objectName())
                        self.checkBoxList.append(self.cb6)
                        self.lineTextList.append(self.s6)
                        self.cb6.clicked.connect(self.on_cb6_clicked)
                        longestText = max(longestText, len(App.ActiveDocument.License))
                        dialogRow += 1
                    if str(key).lower() in LastModifiedDateChkLst:
                        t7 = QtGui.QLabel(value)
                        self.la.addWidget(t7, dialogRow, 0)
                        self.cb7 = QtGui.QCheckBox(updateCb)
                        self.cb7.setObjectName(key)
                        if hasattr(self.cb7, "setCheckState"):
                            self.cb7.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb7.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb7, dialogRow, 1)
                        u7 = QtGui.QLabel("")
                        u7.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u7, dialogRow, 2)
                        self.s7 = QtGui.QLineEdit()
                        self.la.addWidget(self.s7, dialogRow, 3)
                        self.cb7.setObjectName(key)
                        keyLst.append(self.cb7.objectName())
                        self.checkBoxList.append(self.cb7)
                        self.lineTextList.append(self.s7)
                        self.cb7.clicked.connect(self.on_cb7_clicked)
                        dialogRow += 1
                        try:
                            dt = datetime.datetime.strptime(
                                App.ActiveDocument.LastModifiedDate,
                                "%Y-%m-%dT%H:%M:%SZ",
                            )
                            if value == "MM/DD/YYYY":
                                self.s7.setText(
                                    "{0}/{1}/{2}".format(dt.month, dt.day, dt.year)
                                )
                            elif value == "YYYY-MM-DD":
                                self.s7.setText(
                                    "{0}-{1}-{2}".format(dt.year, dt.month, dt.day)
                                )
                            else:
                                self.s7.setText(
                                    "{0}/{1}/{2}".format(dt.day, dt.month, dt.year)
                                )
                        except ValueError:
                            App.Console.PrintLog(
                                "DateTime format not recognised: "
                                + App.ActiveDocument.LastModifiedDate
                                + "\n"
                            )
                            self.s7.setText(
                                "{0}/{1}/{2}".format(
                                    date.today().day,
                                    date.today().month,
                                    date.today().year,
                                )
                            )
                    if str(key).lower() in CreatedDateChkLst:
                        t8 = QtGui.QLabel(value)
                        self.la.addWidget(t8, dialogRow, 0)
                        self.cb8 = QtGui.QCheckBox(updateCb)
                        self.cb8.setObjectName(key)
                        if hasattr(self.cb8, "setCheckState"):
                            self.cb8.setCheckState(QtCore.Qt.Checked)
                        else:
                            self.cb8.setChecked(QtCore.Qt.Checked)
                        self.la.addWidget(self.cb8, dialogRow, 1)
                        u8 = QtGui.QLabel("")
                        u8.setPixmap(pix.scaled(32, 32))
                        self.la.addWidget(u8, dialogRow, 2)
                        self.s8 = QtGui.QLineEdit()
                        self.la.addWidget(self.s8, dialogRow, 3)
                        self.cb8.setObjectName(key)
                        keyLst.append(self.cb8.objectName())
                        self.checkBoxList.append(self.cb8)
                        self.lineTextList.append(self.s8)
                        self.cb8.clicked.connect(self.on_cb8_clicked)
                        dialogRow += 1
                        try:
                            dt = datetime.datetime.strptime(
                                App.ActiveDocument.CreationDate, "%Y-%m-%dT%H:%M:%SZ"
                            )
                            if value == "MM/DD/YYYY":
                                self.s8.setText(
                                    "{0}/{1}/{2}".format(dt.month, dt.day, dt.year)
                                )
                            elif value == "YYYY-MM-DD":
                                self.s8.setText(
                                    "{0}-{1}-{2}".format(dt.year, dt.month, dt.day)
                                )
                            else:
                                self.s8.setText(
                                    "{0}/{1}/{2}".format(dt.day, dt.month, dt.year)
                                )
                        except ValueError:
                            App.Console.PrintLog(
                                "DateTime format not recognised: "
                                + App.ActiveDocument.LastModifiedDate
                                + "\n"
                            )
                            self.s8.setText(
                                "{0}/{1}/{2}".format(
                                    date.today().day,
                                    date.today().month,
                                    date.today().year,
                                )
                            )
                App.Console.PrintLog("\n")
                if len(keyLst) > 1:
                    self.cbAll = QtGui.QCheckBox(
                        QtCore.QT_TRANSLATE_NOOP(
                            "TechDraw_FillTemplateFields",
                            "Update All",
                        )
                    )
                    if hasattr(self.cbAll, "setCheckState"):
                        self.cbAll.setCheckState(QtCore.Qt.Checked)
                    else:
                        self.cbAll.setChecked(QtCore.Qt.Checked)
                    self.la.addWidget(self.cbAll, dialogRow, 1)
                    self.cbAll.clicked.connect(self.on_cbAll_clicked)

                if len(keyLst) > 0:
                    self.okbox = QtGui.QDialogButtonBox(self.dialog)
                    self.okbox.setOrientation(QtCore.Qt.Horizontal)

                    self.okbox.setStandardButtons(
                        QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Ok
                    )
                    self.la.addWidget(self.okbox)
                    QtCore.QObject.connect(
                        self.okbox, QtCore.SIGNAL("accepted()"), self.proceed
                    )
                    QtCore.QObject.connect(
                        self.okbox, QtCore.SIGNAL("rejected()"), self.close
                    )
                    self.okbox.button(QtGui.QDialogButtonBox.Ok).setText("&Ok")
                    self.okbox.button(QtGui.QDialogButtonBox.Cancel).setText("&Cancel")
                    self.button = self.okbox.button(QtGui.QDialogButtonBox.Ok)
                    self.button.setEnabled(True)
                    self.dialog.resize(600 + longestText, dialogRow * 50 + 75)
                    self.dialog.move(400, 200 * (400 / (dialogRow * 50 + 75)))
                    self.dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
                    QtCore.QMetaObject.connectSlotsByName(self.dialog)
                    self.dialog.show()
                    self.dialog.exec_()

                # App.setActiveTransaction("Fill template fields")
                else:
                    msgBox = QtGui.QMessageBox()
                    msgTitle = QtCore.QT_TRANSLATE_NOOP(
                        "Techdraw_FillTemplateFields",
                        "Corresponding template fields missing",
                    )
                    msg = QtCore.QT_TRANSLATE_NOOP(
                        "Techdraw_FillTemplateFields",
                        "There were no corresponding fields found in "
                        + self.page.Label,
                    )
                    msgBox.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
                    msgBox.setText(msg)
                    msgBox.setWindowTitle(msgTitle)
                    msgBox.move(400, 450)
                    msgBox.exec_()

    def on_cbAll_clicked(self):
        if self.cbAll.isChecked():
            for cbEach in self.checkBoxList:
                if hasattr(cbEach, "setCheckState"):
                    cbEach.setCheckState(QtCore.Qt.Checked)
                else:
                    cbEach.setChecked(QtCore.Qt.Checked)
            self.button.setEnabled(True)
        else:
            for cbEach in self.checkBoxList:
                if hasattr(cbEach, "setCheckState"):
                    cbEach.setCheckState(QtCore.Qt.Unchecked)
                else:
                    cbEach.setChecked(QtCore.Qt.Unchecked)
            self.button.setEnabled(False)

    def on_cb1_clicked(self):
        if self.cb1.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb2_clicked(self):
        if self.cb2.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb3_clicked(self):
        if self.cb3.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb4_clicked(self):
        if self.cb4.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb5_clicked(self):
        if self.cb5.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb6_clicked(self):
        if self.cb6.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb7_clicked(self):
        if self.cb7.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def on_cb8_clicked(self):
        if self.cb8.isChecked():
            self.button.setEnabled(True)
        else:
            self.chkEachChkBox()

    def chkEachChkBox(self):
        boolFnd = False
        for cbEach in self.checkBoxList:
            if cbEach.isChecked():
                boolFnd = True
        if boolFnd:
            self.button.setEnabled(True)
        else:
            self.button.setEnabled(False)

    def proceed(self):
        transactionName = QtCore.QT_TRANSLATE_NOOP(
            "Techdraw_FillTemplateFields", "Fill template fields"
        )
        App.setActiveTransaction(transactionName)
        i = 0
        for cb in self.checkBoxList:
            if cb.isChecked():
                self.texts[keyLst[i]] = self.lineTextList[i].text()
            i += 1
        self.page.Template.EditableTexts = self.texts
        App.closeActiveTransaction(False)
        self.close()

        App.closeActiveTransaction()

    def close(self):
        self.dialog.hide()
        keyLst.clear()
        return True



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""GUI for python dependency management."""

import os

import addonmanager_freecad_interface as fci
from addonmanager_python_deps import PythonPackageListModel

from PySideWrapper import QtWidgets


translate = fci.translate


class PythonPackageManagerGui:
    """GUI for managing Python packages"""

    def __init__(self, addons):
        self.dlg = fci.loadUi(
            os.path.join(os.path.dirname(__file__), "PythonDependencyUpdateDialog.ui")
        )
        self.dlg.setObjectName("AddonManager_PythonDependencyUpdateDialog")
        self.model = PythonPackageListModel(addons)
        self.dlg.tableView.setModel(self.model)

        self.dlg.tableView.horizontalHeader().setStretchLastSection(True)
        self.dlg.tableView.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )
        self.dlg.tableView.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )
        self.dlg.tableView.horizontalHeader().setSectionResizeMode(
            2, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )
        self.dlg.tableView.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )

        self.dlg.buttonUpdateAll.clicked.connect(self._update_button_clicked)
        self.model.modelReset.connect(self._model_was_reset)
        self.model.update_complete.connect(self._update_complete)

    def show(self):
        self.dlg.buttonUpdateAll.setEnabled(False)
        self.dlg.updateInProgressLabel.show()
        self.model.reset_package_list()
        self.dlg.labelInstallationPath.setText(self.model.vendor_path)
        self.dlg.exec()

    def _update_button_clicked(self):
        self.dlg.buttonUpdateAll.setEnabled(False)
        self.dlg.updateInProgressLabel.show()
        self.model.update_all_packages()

    def _model_was_reset(self):
        self.dlg.updateInProgressLabel.hide()
        self.dlg.buttonUpdateAll.setEnabled(self.model.updates_are_available())

    def _update_complete(self):
        self.dlg.updateInProgressLabel.hide()
        self.model.reset_package_list()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM mesh netgen task panel for mesh netgen object"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package task_mesh_netgen
#  \ingroup FEM
#  \brief task panel for mesh netgen object

from PySide import QtCore

import FreeCAD
import FreeCADGui

from femmesh import netgentools

from . import base_femlogtaskpanel


class _TaskPanel(base_femlogtaskpanel._BaseLogTaskPanel):
    """
    The TaskPanel for editing References property of
    MeshNetgen objects and creation of new FEM mesh
    """

    def __init__(self, obj):
        super().__init__(obj, netgentools.NetgenTools(obj))

        self.form = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/MeshNetgen.ui"
        )

        self.text_log = self.form.te_output
        self.text_time = self.form.l_time

        self.setup_connections()

    def setup_connections(self):
        super().setup_connections()

        QtCore.QObject.connect(
            self.form.qsb_max_size,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.max_size_changed,
        )
        QtCore.QObject.connect(
            self.form.qsb_min_size,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.min_size_changed,
        )
        QtCore.QObject.connect(
            self.form.dsb_seg_per_edge,
            QtCore.SIGNAL("valueChanged(double)"),
            self.seg_per_edge_changed,
        )
        QtCore.QObject.connect(
            self.form.dsb_curvature_safety,
            QtCore.SIGNAL("valueChanged(double)"),
            self.curvature_safety_changed,
        )
        QtCore.QObject.connect(
            self.form.dsb_growth_rate,
            QtCore.SIGNAL("valueChanged(double)"),
            self.growth_rate_changed,
        )
        QtCore.QObject.connect(
            self.form.ckb_second_order, QtCore.SIGNAL("toggled(bool)"), self.second_order_changed
        )
        QtCore.QObject.connect(
            self.form.cb_fineness,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.fineness_changed,
        )
        QtCore.QObject.connect(
            self.form.pb_get_netgen_version, QtCore.SIGNAL("clicked()"), self.get_version
        )

        self.get_object_params()
        self.set_widgets()

    def get_object_params(self):
        self.min_size = self.obj.MinSize
        self.max_size = self.obj.MaxSize
        self.fineness = self.obj.Fineness
        self.growth_rate = self.obj.GrowthRate
        self.curvature_safety = self.obj.CurvatureSafety
        self.seg_per_edge = self.obj.SegmentsPerEdge
        self.second_order = self.obj.SecondOrder
        self.user_p = self.get_user_fineness_params(self.obj)

    def set_object_params(self):
        self.obj.MinSize = self.min_size
        self.obj.MaxSize = self.max_size
        self.obj.Fineness = self.fineness
        self.obj.GrowthRate = self.growth_rate
        self.obj.CurvatureSafety = self.curvature_safety
        self.obj.SegmentsPerEdge = self.seg_per_edge
        self.obj.SecondOrder = self.second_order

    def set_widgets(self):
        "fills the widgets"
        self.form.qsb_max_size.setProperty("value", self.max_size)
        FreeCADGui.ExpressionBinding(self.form.qsb_max_size).bind(self.obj, "MaxSize")

        self.form.qsb_min_size.setProperty("value", self.min_size)
        FreeCADGui.ExpressionBinding(self.form.qsb_min_size).bind(self.obj, "MinSize")

        self.form.dsb_growth_rate.setProperty("value", self.growth_rate)
        FreeCADGui.ExpressionBinding(self.form.dsb_growth_rate).bind(self.obj, "GrowthRate")

        self.form.dsb_curvature_safety.setProperty("value", self.curvature_safety)
        FreeCADGui.ExpressionBinding(self.form.dsb_curvature_safety).bind(
            self.obj, "CurvatureSafety"
        )

        self.form.dsb_seg_per_edge.setProperty("value", self.seg_per_edge)
        FreeCADGui.ExpressionBinding(self.form.dsb_seg_per_edge).bind(self.obj, "SegmentsPerEdge")

        self.fineness_enum = self.obj.getEnumerationsOfProperty("Fineness")
        index = self.fineness_enum.index(self.fineness)
        self.form.cb_fineness.addItems(self.fineness_enum)
        self.form.cb_fineness.setCurrentIndex(index)

        self.form.ckb_second_order.setChecked(self.second_order)

    def max_size_changed(self, base_quantity_value):
        self.max_size = base_quantity_value

    def min_size_changed(self, base_quantity_value):
        self.min_size = base_quantity_value

    def seg_per_edge_changed(self, value):
        self.seg_per_edge = value

    def curvature_safety_changed(self, value):
        self.curvature_safety = value

    def growth_rate_changed(self, value):
        self.growth_rate = value

    def fineness_changed(self, index):
        self.fineness = self.fineness_enum[index]
        self.obj.Fineness = self.fineness
        if self.fineness == "UserDefined":
            self.form.dsb_seg_per_edge.setEnabled(True)
            self.form.dsb_growth_rate.setEnabled(True)
            self.form.dsb_curvature_safety.setEnabled(True)
            self.form.dsb_seg_per_edge.setProperty("value", self.user_p["segmentsperedge"])
            self.form.dsb_growth_rate.setProperty("value", self.user_p["grading"])
            self.form.dsb_curvature_safety.setProperty("value", self.user_p["curvaturesafety"])
            # set properties
            self.obj.SegmentsPerEdge = self.user_p["segmentsperedge"]
            self.obj.GrowthRate = self.user_p["grading"]
            self.obj.CurvatureSafety = self.user_p["curvaturesafety"]
            self.obj.OptimizationSteps3d = self.user_p["optsteps3d"]
            self.obj.CloseEdgeFactor = self.user_p["closeedgefac"]
        else:
            p = self.obj.Proxy.get_predef_fineness_params(self.fineness)
            self.form.dsb_seg_per_edge.setEnabled(False)
            self.form.dsb_growth_rate.setEnabled(False)
            self.form.dsb_curvature_safety.setEnabled(False)
            self.form.dsb_growth_rate.setProperty("value", p["grading"])
            self.form.dsb_seg_per_edge.setProperty("value", p["segmentsperedge"])
            self.form.dsb_curvature_safety.setProperty("value", p["curvaturesafety"])

    def second_order_changed(self, bool_value):
        self.second_order = bool_value

    def get_user_fineness_params(self, obj):
        # parameters affected by the fineness value. If initial fineness is "UserDefined"
        # use current values otherwise use predefined values
        p = obj.Proxy.get_predef_fineness_params("UserDefined")
        if obj.Fineness == "UserDefined":
            p["curvaturesafety"] = obj.CurvatureSafety
            p["grading"] = obj.GrowthRate
            p["segmentsperedge"] = obj.SegmentsPerEdge
            p["optsteps3d"] = obj.OptimizationSteps3d
            p["closeedgefac"] = obj.CloseEdgeFactor

        return p



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.SetupSheet as PathSetupSheet
import json
import sys

# Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())

from CAMTests.PathTestUtils import PathTestBase


def refstring(string):
    return string.replace(" u'", " '")


class SomeOp(object):
    def __init__(self, obj):
        Path.Log.track(obj, type(obj))
        obj.addProperty("App::PropertyPercent", "StepOver", "Base", "Some help you are")

    @classmethod
    def SetupProperties(cls):
        return ["StepOver"]

    @classmethod
    def Create(cls, name, obj=None, parentJob=None):
        Path.Log.track(name, obj)
        if obj is None:
            obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
        obj.Proxy = SomeOp(obj)
        return obj


class TestPathSetupSheet(PathTestBase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("TestPathSetupSheet")

    def tearDown(self):
        FreeCAD.closeDocument(self.doc.Name)

    def test00(self):
        """Verify SetupSheet templateAttributes"""
        ss = PathSetupSheet.Create().Proxy
        self.doc.recompute()

        attrs = ss.templateAttributes(True, True)

        self.assertEqualLocale(attrs[PathSetupSheet.Template.HorizRapid], "0.00 mm/s")
        self.assertEqualLocale(attrs[PathSetupSheet.Template.VertRapid], "0.00 mm/s")
        self.assertEqualLocale(attrs[PathSetupSheet.Template.SafeHeightOffset], "3.00 mm")
        self.assertEqual(
            attrs[PathSetupSheet.Template.SafeHeightExpression],
            "OpStockZMax+SetupSheet.SafeHeightOffset",
        )
        self.assertEqualLocale(attrs[PathSetupSheet.Template.ClearanceHeightOffset], "5.00 mm")
        self.assertEqual(
            attrs[PathSetupSheet.Template.ClearanceHeightExpression],
            "OpStockZMax+SetupSheet.ClearanceHeightOffset",
        )

    def test01(self):
        """Verify SetupSheet template attributes roundtrip."""
        o1 = PathSetupSheet.Create()
        self.doc.recompute()
        o1.VertRapid = "10 mm/s"
        o1.HorizRapid = "22 mm/s"
        o1.SafeHeightOffset = "18 mm"
        o1.SafeHeightExpression = "Hugo+Olga"
        o1.ClearanceHeightOffset = "23 mm"
        o1.ClearanceHeightExpression = "Peter+Paul"
        o1.StartDepthExpression = "Alpha"
        o1.FinalDepthExpression = "Omega"
        o1.StepDownExpression = "1"

        o2 = PathSetupSheet.Create()
        self.doc.recompute()
        o2.Proxy.setFromTemplate(o1.Proxy.templateAttributes())
        self.doc.recompute()

        # Need to compare the UserString's due to rounding errors depending on the
        # user's unit settings - should have no impact on the validity

        self.assertEqual(o1.VertRapid.UserString, o2.VertRapid.UserString)
        self.assertEqual(o1.HorizRapid.UserString, o2.HorizRapid.UserString)
        self.assertEqual(o1.SafeHeightOffset.UserString, o2.SafeHeightOffset.UserString)
        self.assertEqual(o1.SafeHeightExpression, o2.SafeHeightExpression)
        self.assertEqual(o1.ClearanceHeightOffset.UserString, o2.ClearanceHeightOffset.UserString)
        self.assertEqual(o1.ClearanceHeightExpression, o2.ClearanceHeightExpression)
        self.assertEqual(o1.StartDepthExpression, o2.StartDepthExpression)
        self.assertEqual(o1.FinalDepthExpression, o2.FinalDepthExpression)
        self.assertEqual(o1.StepDownExpression, o2.StepDownExpression)

    def test02(self):
        """Verify default value detection logic."""
        obj = PathSetupSheet.Create()
        ss = obj.Proxy

        self.assertTrue(ss.hasDefaultToolRapids())
        self.assertTrue(ss.hasDefaultOperationHeights())
        self.assertTrue(ss.hasDefaultOperationDepths())

        obj.VertRapid = "1 mm/s"
        self.assertFalse(ss.hasDefaultToolRapids())
        obj.VertRapid = "0 mm/s"
        self.assertTrue(ss.hasDefaultToolRapids())
        obj.HorizRapid = "1 mm/s"
        self.assertFalse(ss.hasDefaultToolRapids())
        obj.HorizRapid = "0 mm/s"
        self.assertTrue(ss.hasDefaultToolRapids())

        obj.SafeHeightOffset = "0 mm"
        self.assertFalse(ss.hasDefaultOperationHeights())
        obj.SafeHeightOffset = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultSafeHeightOffset
        )
        self.assertTrue(ss.hasDefaultOperationHeights())
        obj.ClearanceHeightOffset = "0 mm"
        self.assertFalse(ss.hasDefaultOperationHeights())
        obj.ClearanceHeightOffset = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultClearanceHeightOffset
        )
        self.assertTrue(ss.hasDefaultOperationHeights())

        obj.SafeHeightExpression = "0 mm"
        self.assertFalse(ss.hasDefaultOperationHeights())
        obj.SafeHeightExpression = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultSafeHeightExpression
        )
        self.assertTrue(ss.hasDefaultOperationHeights())
        obj.ClearanceHeightExpression = "0 mm"
        self.assertFalse(ss.hasDefaultOperationHeights())
        obj.ClearanceHeightExpression = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultClearanceHeightExpression
        )
        self.assertTrue(ss.hasDefaultOperationHeights())

        obj.StartDepthExpression = ""
        self.assertFalse(ss.hasDefaultOperationDepths())
        obj.StartDepthExpression = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultStartDepthExpression
        )
        self.assertTrue(ss.hasDefaultOperationDepths())
        obj.FinalDepthExpression = ""
        self.assertFalse(ss.hasDefaultOperationDepths())
        obj.FinalDepthExpression = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultFinalDepthExpression
        )
        self.assertTrue(ss.hasDefaultOperationDepths())
        obj.StepDownExpression = ""
        self.assertFalse(ss.hasDefaultOperationDepths())
        obj.StepDownExpression = ss.decodeAttributeString(
            PathSetupSheet.SetupSheet.DefaultStepDownExpression
        )
        self.assertTrue(ss.hasDefaultOperationDepths())

    def test10(self):
        """Verify template attributes encoding/decoding of floats."""
        ss = PathSetupSheet.Create().Proxy

        self.assertEqual(ss.expressionReference(), "SetupSheet")

        self.assertEqual(str(ss.encodeTemplateAttributes({"00": 13.00})), "{'00': 13.0}")
        self.assertEqual(str(ss.decodeTemplateAttributes({"00": 13.00})), "{'00': 13.0}")

    def test11(self):
        """Verify template attributes encoding/decoding of strings."""
        ss = PathSetupSheet.Create().Proxy

        self.assertEqual(
            str(ss.encodeTemplateAttributes({"00": "hugo"})),
            refstring("{'00': u'hugo'}"),
        )
        self.assertEqual(
            str(ss.encodeTemplateAttributes({"00": "SetupSheet"})),
            refstring("{'00': u'${SetupSheet}'}"),
        )
        self.assertEqual(
            str(ss.encodeTemplateAttributes({"00": "SetupSheet.y"})),
            refstring("{'00': u'${SetupSheet}.y'}"),
        )
        self.assertEqual(
            str(ss.encodeTemplateAttributes({"00": "${hugo}"})),
            refstring("{'00': u'${hugo}'}"),
        )

        self.assertEqual(
            str(ss.decodeTemplateAttributes({"00": "hugo"})),
            refstring("{'00': u'hugo'}"),
        )
        self.assertEqual(
            str(ss.decodeTemplateAttributes({"00": "${SetupSheet}"})),
            refstring("{'00': u'SetupSheet'}"),
        )
        self.assertEqual(
            str(ss.decodeTemplateAttributes({"00": "${SetupSheet}.y"})),
            refstring("{'00': u'SetupSheet.y'}"),
        )
        self.assertEqual(
            str(ss.decodeTemplateAttributes({"00": "${SetupSheet}.y - ${SetupSheet}.z"})),
            refstring("{'00': u'SetupSheet.y - SetupSheet.z'}"),
        )

    def test12(self):
        """Verify template attributes encoding/decoding of dictionaries."""
        ss = PathSetupSheet.Create().Proxy

        self.assertEqual(
            str(ss.encodeTemplateAttributes({"00": {"01": "hugo"}})),
            refstring("{'00': {'01': u'hugo'}}"),
        )
        self.assertEqual(
            str(ss.encodeTemplateAttributes({"00": {"01": "SetupSheet.y - SetupSheet.z"}})),
            refstring("{'00': {'01': u'${SetupSheet}.y - ${SetupSheet}.z'}}"),
        )

        self.assertEqual(
            str(ss.decodeTemplateAttributes({"00": {"01": "hugo"}})),
            refstring("{'00': {'01': u'hugo'}}"),
        )
        self.assertEqual(
            str(ss.decodeTemplateAttributes({"00": {"01": "${SetupSheet}.y - ${SetupSheet}.z"}})),
            refstring("{'00': {'01': u'SetupSheet.y - SetupSheet.z'}}"),
        )

    def test13(self):
        """Verify template attributes encoding/decoding of lists."""
        ss = PathSetupSheet.Create().Proxy

        attrs = {}
        attrs["00"] = "x.SetupSheet"
        attrs["01"] = [{"10": "SetupSheet", "11": "SetupSheet.y"}, {"20": "SetupSheet"}]
        attrs["02"] = [
            {
                "a": [{"b": "SetupSheet"}, {"c": "SetupSheet"}],
                "b": [{"b": "SetupSheet"}],
            }
        ]

        encoded = ss.encodeTemplateAttributes(attrs)
        self.assertEqual(encoded["00"], "x.${SetupSheet}")
        self.assertEqual(len(encoded["01"]), 2)
        self.assertEqual(encoded["01"][0]["10"], "${SetupSheet}")
        self.assertEqual(encoded["01"][0]["11"], "${SetupSheet}.y")
        self.assertEqual(str(encoded["01"][1]), refstring("{'20': u'${SetupSheet}'}"))
        self.assertEqual(len(encoded["02"]), 1)
        self.assertEqual(len(encoded["02"][0]["a"]), 2)
        self.assertEqual(str(encoded["02"][0]["a"][0]), refstring("{'b': u'${SetupSheet}'}"))
        self.assertEqual(str(encoded["02"][0]["a"][1]), refstring("{'c': u'${SetupSheet}'}"))
        self.assertEqual(len(encoded["02"][0]["b"]), 1)
        self.assertEqual(str(encoded["02"][0]["b"][0]), refstring("{'b': u'${SetupSheet}'}"))

        decoded = ss.decodeTemplateAttributes(encoded)
        self.assertEqual(len(decoded), len(attrs))
        self.assertEqual(decoded["00"], attrs["00"])
        self.assertEqual(len(decoded["01"]), len(attrs["01"]))
        self.assertEqual(decoded["01"][0]["10"], attrs["01"][0]["10"])
        self.assertEqual(decoded["01"][0]["11"], attrs["01"][0]["11"])
        self.assertEqual(decoded["01"][1]["20"], attrs["01"][1]["20"])
        self.assertEqual(len(decoded["02"]), len(attrs["02"]))
        self.assertEqual(len(decoded["02"][0]["a"]), len(attrs["02"][0]["a"]))
        self.assertEqual(decoded["02"][0]["a"][0]["b"], attrs["02"][0]["a"][0]["b"])
        self.assertEqual(decoded["02"][0]["a"][1]["c"], attrs["02"][0]["a"][1]["c"])
        self.assertEqual(len(decoded["02"][0]["b"]), len(attrs["02"][0]["b"]))
        self.assertEqual(decoded["02"][0]["b"][0]["b"], attrs["02"][0]["b"][0]["b"])

        # just to be safe ...
        s2 = PathSetupSheet.Create().Proxy
        self.doc.recompute()
        s2.setFromTemplate(ss.templateAttributes())
        self.assertEqual(s2.expressionReference(), "SetupSheet001")
        dec = s2.decodeTemplateAttributes(encoded)
        # pick one
        self.assertEqual(dec["01"][0]["11"], "SetupSheet001.y")

    def test20(self):
        """Verify SetupSheet template op attributes roundtrip."""

        opname = "whoop"

        o1 = PathSetupSheet.Create()

        PathSetupSheet.RegisterOperation(opname, SomeOp.Create, SomeOp.SetupProperties)
        ptt = PathSetupSheet._RegisteredOps[opname].prototype("whoopsy")
        pptt = ptt.getProperty("StepOver")
        pptt.setupProperty(
            o1,
            PathSetupSheet.OpPropertyName(opname, pptt.name),
            PathSetupSheet.OpPropertyGroup(opname),
            75,
        )

        # save setup sheet in json "file"
        attrs = o1.Proxy.templateAttributes(False, False, False, False, [opname])
        encdd = o1.Proxy.encodeTemplateAttributes(attrs)
        j1 = json.dumps({"SetupSheet": encdd}, sort_keys=True, indent=2)

        # restore setup sheet from json "file"
        j2 = json.loads(j1)

        o2 = PathSetupSheet.Create()
        o2.Proxy.setFromTemplate(j2["SetupSheet"])

        op = SomeOp.Create(opname)
        self.assertEqual(op.StepOver, 0)
        o2.Proxy.setOperationProperties(op, opname)
        self.assertEqual(op.StepOver, 75)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Tools for the work with Netgen mesher"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

import numpy as np
import shutil
import sys
import tempfile
from PySide.QtCore import QProcess, QThread, QProcessEnvironment

import FreeCAD
import Fem
from freecad import utils
from femtools.objecttools import ObjectTools


class NetgenTools(ObjectTools):

    # to change order of nodes from netgen to smesh
    order_edge = {
        2: [0, 1, 2],  # seg2
        3: [0, 1, 2],  # seg3
    }
    order_face = {
        3: [0, 1, 2, 3, 4, 5, 6, 7],  # tria3
        6: [0, 1, 2, 5, 3, 4, 6, 7],  # tria6
        4: [0, 1, 2, 3, 4, 5, 6, 7],  # quad4
        8: [0, 1, 2, 3, 4, 7, 5, 6],  # quad8
    }
    order_volume = {
        4: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],  # tetra4
        10: [0, 1, 2, 3, 4, 7, 5, 6, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],  # tetra10
        8: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],  # hexa8
        20: [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 9, 10, 12, 15, 13, 14, 16, 17, 18, 19],  # hexa20
        5: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],  # pyra5
        13: [0, 1, 2, 3, 4, 5, 8, 6, 7, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],  # pyra13
        6: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19],  # penta6
        15: [0, 1, 2, 3, 4, 5, 6, 8, 7, 12, 14, 13, 9, 10, 11, 15, 16, 17, 18, 19],  # penta15
    }

    meshing_step = {
        "AnalyzeGeometry": 1,  # MESHCONST_ANALYSE
        "MeshEdges": 2,  # MESHCONST_MESHEDGES
        "MeshSurface": 3,  # MESHCONST_MESHSURFACE
        "OptimizeSurface": 4,  # MESHCONST_OPTSURFACE
        "MeshVolume": 5,  # MESHCONST_MESHVOLUME
        "OptimizeVolume": 6,  # MESHCONST_OPTVOLUME
    }

    name = "Netgen"
    __param_grp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Netgen")

    def __init__(self, obj):
        super().__init__(obj)
        self.fem_mesh = None
        self.tmpdir = ""
        self.mesh_params = {}

    def write_geom(self):
        if not self.tmpdir:
            self.tmpdir = tempfile.mkdtemp(prefix="fem_")

        global_pla = self.obj.Shape.getGlobalPlacement()
        geom = self.obj.Shape.getPropertyOfGeometry()
        # get partner shape
        geom_trans = geom.transformed(FreeCAD.Placement().Matrix)
        geom_trans.Placement = global_pla
        self.brep_file = self.tmpdir + "/shape.brep"
        self.result_file = self.tmpdir + "/result.npy"
        self.script_file = self.tmpdir + "/code.py"
        geom_trans.exportBrep(self.brep_file)

    def prepare(self):
        self.write_geom()
        self.mesh_params = {
            "brep_file": self.brep_file,
            "threads": self.__param_grp.GetInt("NumOfThreads", QThread.idealThreadCount()),
            "heal": self.obj.HealShape,
            "glue": self.obj.Glue,
            "params": self.get_meshing_parameters(),
            "second_order": self.obj.SecondOrder,
            "second_order_linear": self.obj.SecondOrderLinear,
            "result_file": self.result_file,
            "mesh_region": self.get_mesh_region(),
            "verbosity": self.__param_grp.GetInt("LogVerbosity", 2),
            "zrefine": self.obj.ZRefine,
            "zrefine_size": self.obj.ZRefineSize,
            "zrefine_direction": tuple(self.obj.ZRefineDirection),
        }

        with open(self.script_file, "w") as file:
            file.write(
                self.code.format(
                    kwds=self.mesh_params,
                    order_face=NetgenTools.order_face,
                    order_volume=NetgenTools.order_volume,
                )
            )

    def compute(self):
        env = QProcessEnvironment.systemEnvironment()
        self.process.setProcessEnvironment(env)
        self.process.start(self._get_python_exe(), ["-X", "utf8", "-E", self.script_file])

        return self.process

    code = """
# report Python executable and meshing script
import sys
print("Python interpreter:", sys.executable, flush=True)
print("Meshing script:", *sys.argv, flush=True)

try:
    from netgen import occ, meshing
    import pyngcore as ngcore
    import numpy as np
except ModuleNotFoundError:
    sys.exit("To use FemMesh Netgen, install numpy and Netgen Python bindings")


order_face = {order_face}
order_volume = {order_volume}

def run_netgen(
    brep_file,
    threads,
    heal,
    glue,
    params,
    second_order,
    second_order_linear,
    result_file,
    mesh_region,
    verbosity,
    zrefine,
    zrefine_size,
    zrefine_direction,
):
    geom = occ.OCCGeometry(brep_file)
    ngcore.SetNumThreads(threads)

    shape = geom.shape
    for items, l in mesh_region:
        for t, n in items:
            if t == "Vertex":
                shape.vertices.vertices[n - 1].maxh = l
            elif t == "Edge":
                shape.edges.edges[n - 1].maxh = l
            elif t == "Face":
                shape.faces.faces[n - 1].maxh = l
            elif t == "Solid":
                shape.solids.solids[n - 1].maxh = l

    if zrefine in ["Custom", "Regular"]:
        for sol in shape.solids:
            bottom = sol.faces.Min(zrefine_direction)
            top = sol.faces.Max(zrefine_direction)
            bottom.Identify(top, "bot-top", type=occ.IdentificationType.CLOSESURFACES)

    with ngcore.TaskManager():
        meshing.SetMessageImportance(verbosity)
        geom = occ.OCCGeometry(shape)
        if heal:
            geom.Heal()
        if glue:
            geom.Glue()
        mesh = geom.GenerateMesh(mp=meshing.MeshingParameters(**params))

        if zrefine == "Regular":
            mesh.ZRefine("bot-top", np.arange(zrefine_size[0], 1, zrefine_size[0]))
        elif zrefine == "Custom":
            mesh.ZRefine("bot-top", zrefine_size)

    result = {{
        "coords": [],
        "Edges": [[], []],
        "Faces": [[], []],
        "Volumes": [[], []],
    }}
    groups = {{"Edges": [], "Faces": [], "Solids": []}}

    # save empty data if last step is geometry analysis
    if params["perfstepsend"] == 1:
        np.save(result_file, [result, groups])
        return None

    if second_order:
        if second_order_linear:
            mesh.SetGeometry(None)
        mesh.SecondOrder()

    coords = mesh.Coordinates()

    edges = mesh.Elements1D().NumPy()
    faces = mesh.Elements2D().NumPy()
    volumes = mesh.Elements3D().NumPy()

    nod_edges = edges["nodes"]
    nod_faces = faces["nodes"]
    nod_volumes = volumes["nodes"]

    np_edges = (nod_edges != 0).sum(axis=1).tolist()
    np_faces = faces["np"].tolist()
    np_volumes = volumes["np"].tolist()

    # set smesh node order
    for i in range(faces.size):
        nod_faces[i] = nod_faces[i][order_face[np_faces[i]]]

    for i in range(volumes.size):
        nod_volumes[i] = nod_volumes[i][order_volume[np_volumes[i]]]

    flat_edges = nod_edges[nod_edges != 0].tolist()
    flat_faces = nod_faces[nod_faces != 0].tolist()
    flat_volumes = nod_volumes[nod_volumes != 0].tolist()

    result = {{
        "coords": coords,
        "Edges": [flat_edges, np_edges],
        "Faces": [flat_faces, np_faces],
        "Volumes": [flat_volumes, np_volumes],
    }}

    # create groups
    nb_edges = edges.size
    nb_faces = faces.size
    nb_volumes = volumes.size

    idx_edges = edges["index"]
    idx_faces = faces["index"]
    idx_volumes = volumes["index"]

    for i in np.unique(idx_edges):
        edge_i = (np.nonzero(idx_edges == i)[0] + 1).tolist()
        groups["Edges"].append([i, edge_i])
    for i in np.unique(idx_faces):
        face_i = (np.nonzero(idx_faces == i)[0] + nb_edges + 1).tolist()
        groups["Faces"].append([i, face_i])

    for i in np.unique(idx_volumes):
        volume_i = (np.nonzero(idx_volumes == i)[0] + nb_edges + nb_faces + 1).tolist()
        groups["Solids"].append([i, volume_i])

    np.save(result_file, [result, groups])

# remove traceback
sys.excepthook = lambda type, value, traceback: print(value)

run_netgen(**{kwds})
    """

    def fem_mesh_from_result(self):
        fem_mesh = Fem.FemMesh()

        # load Netgen result
        netgen_result, groups = np.load(self.result_file, allow_pickle=True)

        for node in netgen_result["coords"]:
            fem_mesh.addNode(*node)

        fem_mesh.addEdgeList(*netgen_result["Edges"])
        fem_mesh.addFaceList(*netgen_result["Faces"])
        fem_mesh.addVolumeList(*netgen_result["Volumes"])

        for g in groups["Edges"]:
            grp_id = fem_mesh.addGroup("Edge" + str(g[0]), "Edge")
            fem_mesh.addGroupElements(grp_id, g[1])

        for g in groups["Faces"]:
            grp_id = fem_mesh.addGroup("Face" + str(g[0]), "Face")
            fem_mesh.addGroupElements(grp_id, g[1])

        for g in groups["Solids"]:
            grp_id = fem_mesh.addGroup("Solid" + str(g[0]), "Volume")
            fem_mesh.addGroupElements(grp_id, g[1])

        return fem_mesh

    def update_properties(self):
        self.obj.FemMesh = self.fem_mesh_from_result()

    def get_meshing_parameters(self):
        params = {
            "optimize3d": self.obj.Optimize3d,
            "optimize2d": self.obj.Optimize2d,
            "optsteps3d": self.obj.OptimizationSteps3d,
            "optsteps2d": self.obj.OptimizationSteps2d,
            "opterrpow": self.obj.OptimizationErrorPower,
            "blockfill": self.obj.BlockFill,
            "filldist": self.obj.FillDistance.Value,
            "safety": self.obj.Safety,
            "relinnersafety": self.obj.RelinnerSafety,
            "uselocalh": self.obj.UseLocalH,
            "grading": self.obj.GrowthRate,
            "delaunay": self.obj.Delaunay,
            "delaunay2d": self.obj.Delaunay2d,
            "maxh": self.obj.MaxSize.Value,
            "minh": self.obj.MinSize.Value,
            "startinsurface": self.obj.StartInSurface,
            "checkoverlap": self.obj.CheckOverlap,
            "checkoverlappingboundary": self.obj.CheckOverlappingBoundary,
            "checkchartboundary": self.obj.CheckChartBoundary,
            "curvaturesafety": self.obj.CurvatureSafety,
            "segmentsperedge": self.obj.SegmentsPerEdge,
            "elsizeweight": self.obj.ElementSizeWeight,
            "parthread": self.obj.ParallelMeshing,
            "perfstepsstart": NetgenTools.meshing_step[self.obj.StartStep],
            "perfstepsend": NetgenTools.meshing_step[self.obj.EndStep],
            "giveuptol2d": self.obj.GiveUpTolerance2d,
            "giveuptol": self.obj.GiveUpTolerance,
            "giveuptolopenquads": self.obj.GiveUpToleranceOpenQuads,
            "maxoutersteps": self.obj.MaxOuterSteps,
            "starshapeclass": self.obj.StarShapeClass,
            "baseelnp": self.obj.BaseElementNp,
            "sloppy": self.obj.Sloppy,
            "badellimit": self.obj.BadElementLimit,
            "check_impossible": self.obj.CheckImpossible,
            "only3D_domain_nr": self.obj.Only3dDomainNr,
            "secondorder": self.obj.SecondOrder,
            "elementorder": self.obj.ElementOrder,
            "quad_dominated": self.obj.QuadDominated,
            "try_hexes": self.obj.TryHexes,
            "inverttets": self.obj.InvertTets,
            "inverttrigs": self.obj.InvertTrigs,
            "parallel_meshing": self.obj.ParallelMeshing,
            "nthreads": self.__param_grp.GetInt("NumOfThreads", QThread.idealThreadCount()),
            "closeedgefac": self.obj.CloseEdgeFactor,
        }

        # set specific parameters by fineness
        if self.obj.Fineness == "VeryCoarse":
            params["curvaturesafety"] = 1
            params["segmentsperedge"] = 0.3
            params["grading"] = 0.7
            params["closeedgefac"] = 0.5
            params["optsteps3d"] = 5

        elif self.obj.Fineness == "Coarse":
            params["curvaturesafety"] = 1.5
            params["segmentsperedge"] = 0.5
            params["grading"] = 0.5
            params["closeedgefac"] = 1
            params["optsteps3d"] = 5

        elif self.obj.Fineness == "Moderate":
            params["curvaturesafety"] = 2
            params["segmentsperedge"] = 1
            params["grading"] = 0.3
            params["closeedgefac"] = 2
            params["optsteps3d"] = 5

        elif self.obj.Fineness == "Fine":
            params["curvaturesafety"] = 3
            params["segmentsperedge"] = 2
            params["grading"] = 0.2
            params["closeedgefac"] = 3.5
            params["optsteps3d"] = 5

        elif self.obj.Fineness == "VeryFine":
            params["curvaturesafety"] = 5
            params["segmentsperedge"] = 3
            params["grading"] = 0.1
            params["closeedgefac"] = 5
            params["optsteps3d"] = 5

        return params

    def get_mesh_region(self):
        from Part import Shape as PartShape

        result = []
        for reg in self.obj.MeshRegionList:
            if reg.Suppressed:
                continue
            for s, sub_list in reg.References:
                if s.isDerivedFrom("App::GeoFeature") and isinstance(
                    s.getPropertyOfGeometry(), PartShape
                ):
                    geom = s.getPropertyOfGeometry()
                    sub_obj = [s.getSubObject(_) for _ in sub_list]
                    sub_sh = geom.findSubShape(sub_obj)
                    l = reg.CharacteristicLength.getValueAs("mm").Value
                    result.append((sub_sh, l))
        return result

    @staticmethod
    def _get_python_exe():
        path = NetgenTools.__param_grp.GetString("NetgenPythonPath", "")
        if not path:
            path = utils.get_python_exe()

        return path

    @staticmethod
    def version():
        script = """
try:
    from netgen import config
    print("Netgen:", config.version, flush=True)
except:
    pass
"""
        p = QProcess()
        p.start(NetgenTools._get_python_exe(), ["-E", "-c", script])
        p.waitForFinished()
        info = p.readAll().data().decode()

        return info

    def __del__(self):
        if self.tmpdir:
            shutil.rmtree(self.tmpdir)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Util as PathUtil
from PySide import QtGui, QtCore

from PySide import QtCore, QtGui

__title__ = "CAM UI helper and utility functions"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "A collection of helper and utility functions for the CAM GUI."


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def populateCombobox(form, enumTups, comboBoxesPropertyMap):
    """populateCombobox(form, enumTups, comboBoxesPropertyMap) ... populate comboboxes with translated enumerations
    ** comboBoxesPropertyMap will be unnecessary if UI files use strict combobox naming protocol.
    Args:
        form = UI form
        enumTups = list of (translated_text, data_string) tuples
        comboBoxesPropertyMap = list of (translated_text, data_string) tuples
    """
    Path.Log.track(enumTups)

    # Load appropriate enumerations in each combobox
    for cb, prop in comboBoxesPropertyMap:
        box = getattr(form, cb)  # Get the combobox
        box.clear()  # clear the combobox
        for text, data in enumTups[prop]:  #  load enumerations
            box.addItem(text, data)


def updateInputField(obj, prop, widget, onBeforeChange=None):
    """updateInputField(obj, prop, widget) ... update obj's property prop with the value of widget.
    The property's value is only assigned if the new value differs from the current value.
    This prevents onChanged notifications where the value didn't actually change.
    Gui::InputField and Gui::QuantitySpinBox widgets are supported - and the property can
    be of type Quantity or Float.
    If onBeforeChange is specified it is called before a new value is assigned to the property.
    Returns True if a new value was assigned, False otherwise (new value is the same as the current).
    """
    value = widget.property("rawValue")
    Path.Log.track("value: {}".format(value))
    attr = PathUtil.getProperty(obj, prop)
    attrValue = attr.Value if hasattr(attr, "Value") else attr

    isDiff = False
    if not Path.Geom.isRoughly(attrValue, value):
        isDiff = True
    else:
        if hasattr(obj, "ExpressionEngine"):
            exprSet = False
            for prp, expr in obj.ExpressionEngine:
                if prp == prop:
                    exprSet = True
                    Path.Log.debug('prop = "expression": {} = "{}"'.format(prp, expr))
                    value = FreeCAD.Units.Quantity(obj.evalExpression(expr)).Value
                    if not Path.Geom.isRoughly(attrValue, value):
                        isDiff = True
                    break
            if exprSet:
                widget.setReadOnly(True)
                widget.setProperty("exprSet", "true")
                widget.style().unpolish(widget)
                widget.ensurePolished()
            else:
                widget.setReadOnly(False)
                widget.setProperty("exprSet", "false")
                widget.style().unpolish(widget)
                widget.ensurePolished()
            widget.update()

    if isDiff:
        Path.Log.debug("updateInputField(%s, %s): %.2f -> %.2f" % (obj.Label, prop, attr, value))
        if onBeforeChange:
            onBeforeChange(obj)
        PathUtil.setProperty(obj, prop, value)
        return True

    return False


class QuantitySpinBox(QtCore.QObject):
    """Controller class to interface a Gui::QuantitySpinBox.
    The spin box gets bound to a given property and supports update in both directions.
    QuatitySpinBox(widget, obj, prop, onBeforeChange=None)
            widget ... expected to be reference to a Gui::QuantitySpinBox
            obj    ... document object
            prop   ... canonical name of the (sub-) property
            onBeforeChange ... an optional callback being executed before the value of the property is changed
    """

    def __init__(self, widget, obj, prop, onBeforeChange=None):
        super().__init__()
        Path.Log.track(widget)
        self.widget = widget
        self.onBeforeChange = onBeforeChange
        self.prop = None
        self.obj = obj
        self.lastWidgetText = self.widget.text()
        self.attachTo(obj, prop)
        self.widget.installEventFilter(self)
        # Connect local class method as slot
        self.widget.textChanged.connect(self.onWidgetValueChanged)
        # Connect to showFormulaDialog signal to track dialog open/close
        try:
            self.widget.showFormulaDialog.connect(self.onFormulaDialogStateChanged)
        except AttributeError:
            # Widget may not have this signal
            pass

    def onFormulaDialogStateChanged(self, isOpen):
        """
        Slot called when the formula dialog is opened or closed.
        When the dialog closes (isOpen=False), refresh the widget.
        """
        if not isOpen:
            # Dialog has closed, update the widget to reflect any changes
            self.updateWidget()

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Type.FocusIn:
            self.updateWidget()
        return False

    def onWidgetValueChanged(self):
        """
        Slot method for determining if a change
        in widget value is a result of an expression edit, or a simple spinbox change.
        If the former, emit a manual `editingFinished` signal because the Expression editor
        window returned a value to the base widget, leaving it in read-only mode,
        and finishing the editing of the value. Otherwise, due nothing if the value
        has not changed, or there is no active expression for the property.
        If the user closes the Expression editor to cancel the edit, the value will not
        be changed, and this manual signal will not be emitted."""
        if self._hasExpression() and self.widget.text() != self.lastWidgetText:
            self.widget.editingFinished.emit()

    def attachTo(self, obj, prop=None):
        """use an existing editor for the given object and property"""
        Path.Log.track(self.prop, prop)
        self.obj = obj
        self.prop = prop
        if obj and prop:
            attr = PathUtil.getProperty(obj, prop)
            if attr is not None:
                if hasattr(attr, "Value"):
                    self.widget.setProperty("unit", attr.getUserPreferred()[2])
                self.widget.setProperty("binding", "%s.%s" % (obj.Name, prop))
                self.valid = True
            else:
                Path.Log.warning("Cannot find property {} of {}".format(prop, obj.Label))
                self.valid = False
        else:
            self.valid = False

    def expression(self):
        """returns the expression if one is bound to the property"""
        Path.Log.track(self.prop, self.valid)
        if self.valid:
            return self.widget.property("expression")
        return ""

    def setMinimum(self, quantity):
        """set the minimum"""
        Path.Log.track(self.prop, self.valid)
        if self.valid:
            value = quantity.Value if hasattr(quantity, "Value") else quantity
            self.widget.setProperty("setMinimum", value)

    def updateWidget(self, quantity=None):
        """
        update the display value of the spin box.
        If no value is provided the value of the bound property is used.
        quantity can be of type Quantity or Float."""
        Path.Log.track(self.prop, self.valid, quantity)

        if self.valid:
            expr = self._hasExpression()
            if quantity is None:
                if expr:
                    quantity = FreeCAD.Units.Quantity(self.obj.evalExpression(expr))
                else:
                    quantity = PathUtil.getProperty(self.obj, self.prop)
            value = quantity.Value if hasattr(quantity, "Value") else quantity
            self.widget.setProperty("rawValue", value)
            self.lastWidgetText = self.widget.text()  # update last widget value
            if expr:
                self.widget.setReadOnly(True)
                self.widget.setProperty("exprSet", "true")
                self.widget.style().unpolish(self.widget)
                self.widget.ensurePolished()
            else:
                self.widget.setReadOnly(False)
                self.widget.setProperty("exprSet", "false")
                self.widget.style().unpolish(self.widget)
                self.widget.ensurePolished()

    def updateProperty(self):
        """updateProperty() ... update the bound property with the value from the spin box"""
        Path.Log.track(self.prop, self.valid)
        if self.valid:
            return updateInputField(self.obj, self.prop, self.widget, self.onBeforeChange)
        return None

    def _hasExpression(self):
        for prop, exp in self.obj.ExpressionEngine:
            if prop == self.prop:
                return exp
        return None


class PropertyComboBox(QtCore.QObject):
    """Base controller class for properties represented as QComboBox."""

    def __init__(self, widget, obj, prop, onBeforeChange=None):
        super().__init__()
        Path.Log.track(widget)
        self.widget = widget
        self.onBeforeChange = onBeforeChange
        self.prop = None
        self.obj = obj
        self.valid = False
        self.attachTo(obj, prop)
        self.widget.currentIndexChanged.connect(self.updateProperty)

    def attachTo(self, obj, prop=None):
        """use an existing editor for the given object and property"""
        Path.Log.track(self.prop, prop)
        self.obj = obj
        self.prop = prop
        if obj and prop:
            attr = PathUtil.getProperty(obj, prop)
            if attr is not None:
                self.valid = True
                self._populateComboBox()
                self.updateWidget()
            else:
                Path.Log.warning("Cannot find property {} of {}".format(prop, obj.Label))
                self.valid = False
        else:
            self.valid = False

    def _populateComboBox(self):
        """To be implemented by subclasses"""
        raise NotImplementedError

    def updateWidget(self, value=None):
        """update the display value of the combo box."""
        Path.Log.track(self.prop, self.valid, value)
        if self.valid:
            if value is None:
                value = PathUtil.getProperty(self.obj, self.prop)
            index = (
                self.widget.findData(value)
                if hasattr(self.widget, "findData")
                else self.widget.findText(str(value))
            )
            if index >= 0:
                self.widget.setCurrentIndex(index)

    def updateProperty(self):
        """update the bound property with the value from the combo box"""
        Path.Log.track(self.prop, self.valid)
        if self.valid and self.prop:
            if self.onBeforeChange:
                self.onBeforeChange()

            current_value = PathUtil.getProperty(self.obj, self.prop)
            new_value = (
                self.widget.currentData()
                if hasattr(self.widget, "currentData")
                else self.widget.currentText()
            )

            if str(new_value) != str(current_value):
                setattr(self.obj, self.prop, new_value)
                return True
        return False


class IntegerSpinBox(QtCore.QObject):
    """Controller class for integer properties represented as QSpinBox.
    IntegerSpinBox(widget, obj, prop, onBeforeChange=None)
            widget ... expected to be reference to a QSpinBox
            obj    ... document object
            prop   ... canonical name of the (sub-) property
            onBeforeChange ... optional callback before property change
    """

    def __init__(self, widget, obj, prop, onBeforeChange=None):
        super().__init__()
        self.widget = widget
        self.onBeforeChange = onBeforeChange
        self.prop = None
        self.obj = obj
        self.valid = False

        # Configure spin box defaults
        self.widget.setMinimum(-2147483647)  # Qt's minimum for spin boxes
        self.widget.setMaximum(2147483647)  # Qt's maximum for spin boxes

        self.attachTo(obj, prop)
        self.widget.valueChanged.connect(self.updateProperty)

    def attachTo(self, obj, prop=None):
        """bind to the given object and property"""
        self.obj = obj
        self.prop = prop
        if obj and prop:
            try:
                prop_value = PathUtil.getProperty(obj, prop)
                if prop_value is not None:
                    self.valid = True
                    self.updateWidget()
                else:
                    Path.Log.warning(f"Cannot get value for property {prop} of {obj.Label}")
                    self.valid = False
            except Exception as e:
                Path.Log.error(f"Error attaching to property {prop}: {str(e)}")
                self.valid = False
        else:
            self.valid = False

    def updateWidget(self, value=None):
        """update the spin box value"""
        if self.valid:
            try:
                if value is None:
                    value = PathUtil.getProperty(self.obj, self.prop)

                # Handle both direct values and Quantity objects
                if hasattr(value, "Value"):  # For Quantity properties
                    value = int(value.Value)

                self.widget.setValue(int(value))
            except Exception as e:
                Path.Log.error(f"Error updating spin box: {str(e)}")

    def updateProperty(self):
        """update the bound property with the spin box value"""
        if self.valid and self.prop:
            if self.onBeforeChange:
                self.onBeforeChange()

            new_value = self.widget.value()
            current_value = PathUtil.getProperty(self.obj, self.prop)

            # Handle Quantity properties
            if hasattr(current_value, "Value"):
                if new_value != current_value.Value:
                    current_value.Value = new_value
                    return True
            elif new_value != current_value:
                setattr(self.obj, self.prop, new_value)
                return True
        return False

    def setRange(self, min_val, max_val):
        """set minimum and maximum values"""
        self.widget.setMinimum(min_val)
        self.widget.setMaximum(max_val)

    def setSingleStep(self, step):
        """setSingleStep(step) ... set the step size"""
        self.widget.setSingleStep(step)


class BooleanComboBox(PropertyComboBox):
    """Controller class for boolean properties represented as QComboBox."""

    def _populateComboBox(self):
        self.widget.clear()
        self.widget.addItem("True", True)
        self.widget.addItem("False", False)


class EnumerationComboBox(PropertyComboBox):
    """Controller class for enumeration properties represented as QComboBox."""

    def _populateComboBox(self):
        self.widget.clear()
        enums = self.obj.getEnumerationsOfProperty(self.prop)
        for item in enums:
            self.widget.addItem(item, item)


class PropertyLabel(QtCore.QObject):
    """Controller class for read-only property display as QLabel."""

    def __init__(self, widget, obj, prop, onBeforeChange=None):
        super().__init__()
        self.widget = widget
        self.obj = obj
        self.prop = prop
        self.valid = False
        self.attachTo(obj, prop)

    def attachTo(self, obj, prop=None):
        """bind to the given object and property"""
        self.obj = obj
        self.prop = prop
        if obj and prop:
            attr = PathUtil.getProperty(obj, prop)
            if attr is not None:
                self.valid = True
                self.updateWidget()
            else:
                Path.Log.warning(f"Cannot find property {prop} of {obj.Label}")
                self.valid = False
        else:
            self.valid = False

    def updateWidget(self, value=None):
        """update the label text"""
        if self.valid:
            if value is None:
                value = PathUtil.getProperty(self.obj, self.prop)
            self.widget.setText(str(value))


def getDocNode():
    doc = FreeCADGui.ActiveDocument.Document.Name
    tws = FreeCADGui.getMainWindow().findChildren(QtGui.QTreeWidget)

    for tw in tws:
        if tw.topLevelItemCount() != 1 or tw.topLevelItem(0).text(0) != "Application":
            continue
        toptree = tw.topLevelItem(0)
        for i in range(0, toptree.childCount()):
            docitem = toptree.child(i)
            if docitem.text(0) == doc:
                return docitem
    return None


def disableItem(item):
    Dragflag = QtCore.Qt.ItemFlag.ItemIsDragEnabled
    Dropflag = QtCore.Qt.ItemFlag.ItemIsDropEnabled
    item.setFlags(item.flags() & ~Dragflag)
    item.setFlags(item.flags() & ~Dropflag)
    for idx in range(0, item.childCount()):
        disableItem(item.child(idx))


def findItem(docitem, objname):
    print(docitem.text(0))
    for i in range(0, docitem.childCount()):
        if docitem.child(i).text(0) == objname:
            return docitem.child(i)
        res = findItem(docitem.child(i), objname)
        if res:
            return res
    return None



# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2022 - 2025 Larry Woestman <LarryWoestman2@gmail.com>   *
# *   Copyright (c) 2025 Alan Grover <awgrover@gmail.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""FreeCAD CAM post-processor for Shopbot native opensbb, in the "refactored" style"""

import re
import argparse
from copy import copy
import operator
import math
import time

from typing import Any

from Path.Post.Processor import PostProcessor
import Path.Post.UtilsArguments as PostUtilsArguments
import Path.Post.UtilsExport as PostUtilsExport
import Path.Post.UtilsParse as PostUtilsParse

import Path
import FreeCAD
import FreeCADGui

translate = FreeCAD.Qt.translate

DEBUG = False
if DEBUG:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

#
# Define some types that are used throughout this file.
#
Defaults = dict[str | bool]
FormatHelp = str
GCodeOrNone = [str | None]
GCodeSections = list[tuple[str, GCodeOrNone]]
Parser = argparse.ArgumentParser
ParserArgs = [None | str | argparse.Namespace]
Postables = [list | list[tuple[str, list]]]
Section = tuple[str, list]
Sublist = list
Units = str
Values = dict[str, Any]
Visible = dict[str, bool]
nl = "\n"  # particularly useful in a f-string
PreventGuiTimeout = 1  # seconds

SpeedPrecisions = {"G20": 2, "G21": 1}  # for in/sec, for mm/sec
MinSpeeds = {"G20": 0.05, "G21": 1.3}  # for in/sec, for mm/sec


class Opensbp(PostProcessor):
    """For ShopBot (or other opensbp controllers), this is a CAM postprocessor.
    We have to translate gcode to opensbp, so the output is NOT gcode.
    Typically, you'd want a .sbp file extension.

    Special behaviors:
    * Many defaults have changed: read the tooltip for the Arguments
    * This always tests the ShopBot app/machine units (see --inches --metric), and on the machine will give an error and exit if in the wrong units. No more mismatches.
    * G54 (fixture/coordinate-system) is accepted, but is a noop (because it's the default in Operations). Others (G55 etc) are not-accepted.
    * Note the default of --tool-change == False. Use that for manual tool change, but you only get 1 tool per file. True means call C9. Note: C9 seems to not work when there is no auto-tool-changer (corrupts units and erases your settings file!)
    * This does arc gcodes (G02 and G03), so helical operations should work.
    * This does Probe (G38.2), defaults to the same folder as the .sbp, and .txt as the extension for output filename
    * Relative movement (G91) is not supported.
    * opensbp commands can be in-lined via a comment:
            (MC_RUN_COMMAND: <opensbpcommand>)
    * By default, this post-processor fails if it sees a gcode it doesn't understand. but see --abort-on-unknown and --allow-unknown
    * Do opensbp prompt (dialog-box) behavior like this:
        (this comment becomes a dialog-box if followed by M00)
        M00
    * --modal only applies to G00 and G01. Notably NOT to arcs.
    * Note that --preamble occurs before the machine-units are set, could be a problem! (not under our control)
    # preamble/postamble are G-Code.
    * Most optimizations assume that the instruction stream isn't interrupted with manually added commands (e.g. assumes the xyz position from the previous instructions). So, don't stop, and insert commands if you've used any of the optimizations.
    * This always sets the move & jog speeds (feed and rapid) at the beginning of each operation, by fetching the Tool speeds. If rapid-speeds are 0.0, it will not set the jog-speed, so will use the machine setting.
    * For spindle speed changes, your ShopBot C6 "cut" command should be configured for having-a-speed-control-module, or for not having. It will prompt you if configured for not-having the module.
    * The A and B axis are not supported
    """

    def __init__(
        self,
        job,
        tooltip=translate("CAM", "see property below"),
        tooltipargs=[""],
        units="Metric",
    ) -> None:
        super().__init__(
            job=job,
            tooltip=tooltip,
            tooltipargs=tooltipargs,
            units=units,
        )
        self.reinitialize()
        Path.Log.debug("Refactored opensbp post processor initialized.")

    def reinitialize(self) -> None:
        """Initialize or reinitialize the 'core' data structures for the postprocessor."""
        #
        # This is also used to reinitialize the data structures between tests.
        #
        self.values: Values = {}
        self.init_values(self.values)
        # debug_str = "\n".join(f"{k}:{self.values[k]}" for k in sorted(self.values.keys())); print(f"### .values\n{debug_str}")

        self.argument_defaults: Defaults = {}
        self.init_argument_defaults(self.argument_defaults)
        Path.Log.debug(f".argument_defaults\n{self.argument_defaults}")
        self.arguments_visible: Visible = {}
        self.init_arguments_visible(self.arguments_visible)
        self.parser: Parser = self.init_arguments(
            self.values, self.argument_defaults, self.arguments_visible
        )
        self.arguments = None  # the parser.parse_args result

        #
        # Create another parser just to get a list of all possible arguments
        # that may be output using --output_all_arguments.
        #
        self.all_arguments_visible: Visible = {}
        for k in iter(self.arguments_visible):
            self.all_arguments_visible[k] = True
        self.all_visible: Parser = self.init_arguments(
            self.values, self.argument_defaults, self.all_arguments_visible
        )

    def init_values(self, values: Values) -> None:
        """Initialize values that are used throughout the postprocessor."""
        #
        PostUtilsArguments.init_shared_values(values)
        #
        # Set any values here that need to override the default values set
        # in the init_shared_values routine.
        #

        # Used in the argparser code as the "name" of the postprocessor program.
        values["MACHINE_NAME"] = "opensbp"

        values.update(
            {
                "DRILL_CYCLES_TO_TRANSLATE": ["G73", "G81", "G82", "G83", "G85"],
                "ENABLE_COOLANT": False,
                "ENABLE_MACHINE_SPECIFIC_COMMANDS": True,
                "LINE_INCREMENT": 1,
                "OUTPUT_PATH_LABELS": True,
                "OUTPUT_TOOL_CHANGE": False,
                "OUTPUT_BLANK_LINES": False,
                # 'PARAMETER_ORDER' : don't care about order, we aren't gcode
                "POSTAMBLE": "",
                "POSTPROCESSOR_FILE_NAME": __name__,
                "PREAMBLE": "",
                "SPINDLE_WAIT": 3,  # for auto case
                "STOP_SPINDLE_FOR_TOOL_CHANGE": True,
                "SUPPRESS_COMMANDS": [
                    "G54"
                ],  # we don't have coord-systems (yet) # G99,G98,G80 added automatically
                # 'TOOL_CHANGE' : we have to generate this dynamically
                "TRANSLATE_DRILL_CYCLES": True,
                "UNITS": self._units,
                "UNIT_SPEED_FORMAT": "mm/s",
                "USE_TLO": False,
                "line_number": 1,
                "ALLOW_UNKNOWN": [],  # --allow-unknown
                "last_command": None,
                "first_probe": True,
                "SPEED_PRECISION": 2,  # updated at process_arguments
                "MIN_SPEED": 0.05,  # updated at process_arguments
                # these seem safe defaults
                "MODAL": True,  # --modal
                "OUTPUT_DOUBLES": False,  # --axis-modal
            }
        )
        # FIXME: should be done by PostProcessor, isn't there yet in 1.0.
        if "G38.2" not in self.values["MOTION_COMMANDS"]:
            self.values["MOTION_COMMANDS"].append("G38.2")

    def init_argument_defaults(self, argument_defaults: Defaults) -> None:
        """Initialize which arguments (in a pair) are shown as the default argument."""
        #
        # Modify which argument to show as the default in flag-type arguments here.
        # If the value is True, the first argument will be shown as the default.
        # If the value is False, the second argument will be shown as the default.
        #
        # For example, if you want to show Metric mode as the default, use:
        #   argument_defaults["metric_inch"] = True
        #
        # If you want to show that "Don't pop up editor for writing output" is
        # the default, use:
        #   argument_defaults["show-editor"] = False.
        PostUtilsArguments.init_argument_defaults(argument_defaults)
        #
        # Note:  You also need to modify the corresponding entries in the "values" hash
        #        to actually make the default value(s) change to match.
        #

        # argument_defaults['metric_inches'] = True if From doc
        argument_defaults["wait-for-spindle"] = self.values["SPINDLE_WAIT"]
        defaults = {
            "tlo": False,
            "translate_drill": True,
            "tool_change": False,
            "modal": True,
            "no_modal": False,
            "axis_modal": True,
            "no_axis_modal": False,
        }
        for arg, v in defaults.items():
            argument_defaults[arg] = v

    def init_arguments_visible(self, arguments_visible: Visible) -> None:
        """Initialize which argument pairs are visible in TOOLTIP_ARGS."""
        PostUtilsArguments.init_arguments_visible(arguments_visible)
        #
        # Modify the visibility of any arguments from the defaults here.
        #
        arguments_visible.update(
            {
                k: False
                for k in (
                    "bcnc",
                    "tlo",
                    "translate_drill",
                    "tool-change",
                    "line-numbers",
                    "gcode-comments",  # FIXME: doesn't hide "our" arguments
                    "axis-modal",
                    "modal",
                )
            }
        )
        arguments_visible.update(
            {
                k: True
                for k in (
                    "return_to",
                    "wait-for-spindle",
                )
            }
        )

    def init_arguments(
        self,
        values: Values,
        argument_defaults: Defaults,
        arguments_visible: Visible,
    ) -> Parser:
        """Initialize the shared argument definitions."""

        _parser: Parser = PostUtilsArguments.init_shared_arguments(
            values, argument_defaults, arguments_visible
        )
        #
        # Add any argument definitions that are not shared with all other
        # postprocessors here.
        #
        _parser.add_argument(
            "--o1",
            action="store_true",
            help="--no-comments --no-header",
        )

        # _parser.add_argument(
        #    # this should probably be True for most shopbot installations
        #    "--ab-is-distance", action="store_true", help="A & B axis are distances, default=degrees"
        # )
        """ # FIXME: not-implemented
        _parser.add_argument(
            "--filter",
            "--filters",
            help="a ',' list of filters in FreeCAD.getUserMacroDir()/post to run on the gcode of each Path object, before we see it (i.e. cleanups). A class of same (camelcase) name as file, __init__(self,objectslist, filename, argstring), .filter(eachpathobj, its-.Commands) -> gcode",
        )
        """
        _parser.add_argument(
            "--abort-on-unknown",
            action=argparse.BooleanOptionalAction,
            help="Generate an error and fail if an unknown gcode is seen. default=True",
            default=True,
        )
        _parser.add_argument(
            "--allow-unknown",
            help="if --abort-on-unknown, allow these gcodes, but change them to a comment. E.g. --allow-unknown G55,G56. Always includes G54,G99,G98,G80",
        )
        _parser.add_argument(
            "--native-rapid-fallback",
            action=argparse.BooleanOptionalAction,
            help="Use machine's rapid speeds if ToolController rapid speeds are zero, default=true. --no-native-rapid-fallback causes error on zeros.",
            default=True,
        )
        _parser.add_argument(
            "--gcode-comments",
            action=argparse.BooleanOptionalAction,
            # help="Add the original gcode as a comment, for debugging",
            help=argparse.SUPPRESS,
            default=False,
        )

        return _parser

    def process_arguments(self) -> tuple[bool, ParserArgs]:
        """Process any arguments to the postprocessor."""
        #
        # This function is separated out to make it easier to inherit from this postprocessor
        #
        args: ParserArgs
        flag: bool

        (flag, args) = PostUtilsArguments.process_shared_arguments(
            self.values, self.parser, self._job.PostProcessorArgs, self.all_visible, "-"
        )
        #
        # If the flag is True, then all of the arguments should be processed normally.
        #
        if flag:
            #
            # Process any additional arguments here.
            #
            #
            # Update any variables that might have been modified while processing the arguments.
            #
            self._units = self.values["UNITS"]
            self.values["UNIT_SPEED_FORMAT"] = (
                "mm/s" if self.values["UNIT_FORMAT"] == "mm" else "in/s"
            )
            self.values["SPEED_PRECISION"] = SpeedPrecisions[self.values["UNITS"]]
            self.values["MIN_SPEED"] = MinSpeeds[self.values["UNITS"]]

            if args.allow_unknown:
                self.values["SUPPRESS_COMMANDS"].extend(
                    # and canonicalize
                    [
                        re.sub(r"^([A-Z])(\d(\.|$))", r"\g<1>0\2", g)
                        for g in args.allow_unknown.split(",")
                    ]
                )

            # too late for .values, so do them by hand
            arg_sets = {
                "o1": {
                    "cli": {
                        "comments": False,
                        "no_comments": True,
                        "no_header": True,
                        "header": False,
                    },
                    "values": {"OUTPUT_HEADER": False, "OUTPUT_COMMENTS": False},
                },
            }

            for opt in ("o1", "o2", "o3"):
                if getattr(args, opt, None):
                    cli_set = arg_sets[opt]["cli"]
                    value_set = arg_sets[opt]["values"]
                    for arg_name, value in cli_set.items():
                        Path.Log.debug(f"--{opt} set {arg_name}={value}")
                        setattr(args, arg_name, value)
                    for value_key, value in value_set.items():
                        self.values[value_key] = value

        #
        # If the flag is False, then args is either None (indicating an error while
        # processing the arguments) or a string containing the argument list formatted
        # for output.  Either way the calling routine will need to handle the args value.
        #
        return (flag, args)

    def process_postables(self) -> GCodeSections:
        """Postprocess the 'postables' in the job to g code sections."""
        #
        # This function is separated out to make it easier to inherit from this postprocessor.
        #
        gcode: GCodeOrNone
        g_code_sections: GCodeSections
        partname: str
        postables: Postables
        section: Section
        sublist: Sublist

        postables = self._buildPostList()

        Path.Log.debug(f"postables count: {len(postables)}")

        g_code_sections = []
        # a section seems to be for one output file of gcode (a "split")
        for _, section in enumerate(postables):
            # partname is "allitems" if not splitting, and the operation/etc. if splitting
            # sublist is each "phase" of the operation, like Fixture, ToolController, Operation
            partname, sublist = section

            # We have to lie about self.values to get some expanded g-code that we rely on
            # e.g. we rely on comments
            was_values = copy(self.values)
            self.values.update(
                {
                    "MODAL": False,  # if true, this would elide the gcode "command", which screws us up. so, we do it later.
                    "OUTPUT_COMMENTS": True,  # we use this to detect operation and...
                    "OUTPUT_TOOL_CHANGE": True,  # we need this to ensure speeds
                    "SHOW_MACHINE_UNITS": True,  # we have to set the machine
                    "AXIS_PRECISION": 5,  # we do calculations, so more precision to prevent rounding errors
                    "FEED_PRECISION": 5,
                    "SPINDLE_WAIT": 0,  # we'll do the logic in the right place later
                }
            )
            try:
                # We get back a processed (expanded) set of gcode
                # things like pre/post amble, canned-drill expansion, coolant on/off
                # We could note where each `partname` begins, if we need that
                #   but we get no access to where each `sublist` starts/ends
                # `gcode` is a multi-line string
                gcode = PostUtilsExport.export_common(self.values, sublist, "-")
            finally:
                self.values = was_values

            # print(f"-gcode expanded-\n{gcode}--")

            # ToOpenSBP will modify our .values, and not restore them
            was_values = copy(self.values)
            try:
                # Treat each `sublist` as a unit/file, so new instance
                native = ToOpenSBP(self).translate(gcode)
            finally:
                self.values = was_values

            g_code_sections.append((partname, native))

        return g_code_sections

    def export(self) -> GCodeSections:
        """Process the parser arguments, then postprocess the 'postables'."""
        args: ParserArgs
        flag: bool

        Path.Log.debug("Exporting the job")
        self.reinitialize()

        (flag, args) = self.process_arguments()
        #
        # If the flag is True, then continue postprocessing the 'postables'
        #
        if flag:
            self.arguments = args
            Path.Log.debug(f"cli args {self._job.PostProcessorArgs}")
            Path.Log.debug(f"argparsed {args}")
            return self.process_postables()
        #
        # The flag is False meaning something unusual happened.
        #
        # If args is None then there was an error during argument processing.
        #
        if args is None:
            return None
        #
        # Otherwise args will contain the argument list formatted for output
        # instead of the "usual" gcode.
        #
        return [("allitems", args)]  # type: ignore

    @property
    def tooltip(self):
        return self.__doc__

    @property
    def tooltipArgs(self) -> FormatHelp:
        return self.parser.format_help()

    @property
    def units(self) -> Units:
        return self._units


def gcode(*commands):
    """Only for use in class ToOpenSBP, for methods.
    Decorator to add the function to the translate-map.
        `commands` is list of Gcodes, i.e. "G54", "G55",
        specify two digit gcode: e.g. "G01", "M06"
    """

    def gcode(func):
        # we really just want to make a map with it
        # But, we want the map to belong to the class
        # which doesn't exist at "use" time
        # so, annotate the method
        # and collect/insert after the class
        func._gcode = []
        for c in commands:
            func._gcode.append(c)
            canonical = re.sub(r"^([A-Z])(\d(\.|$))", r"\g<1>0\2", c)
            if canonical != c:
                raise Exception(f"Internal: Must be a 2 digit gcode (e.g. G00): @gcode('{c}')")

        # null wrapper
        # FIXME: I think if we wrap it, then do obj.func(args), we'll get proper inhertance dispatch
        return func

    return gcode


def gcode_insertmap():
    # see gcode() above. this inserts the map
    for attr in ToOpenSBP.__dict__.values():
        if callable(attr) and hasattr(attr, "_gcode"):
            for g in attr._gcode:
                ToOpenSBP.DispatchMap[g] = attr


class ToOpenSBP:
    """Translate gcode to opensbp"""

    PositionAxis = "XYZAB"  # though AB is barely supported here

    DispatchMap = {}

    def __init__(self, postprocessor: Opensbp):
        # we will use specific features (.values[x] and arguments from the opensbp post)
        self.post = postprocessor

        # xyzf etc state
        self.current_location = {p: None for p in self.post.values["PARAMETER_ORDER"]}
        # for our speed-modal
        self.current_location["ms"] = ["", ""]
        self.current_location["js"] = ["", ""]
        self.end_location = [None for x in self.PositionAxis]

        self.set_units = None  # flag and memory of the first time we see a G20/G21 set-units
        self._postfix = []  # balancing things to add to end
        self.last_gui_update = 0
        self.first_tool = True
        self.first_no_F = True

    def translate(self, gcode):
        """Entry point, returns the translated contents, e.g. opensbp lines.
        We rely on self.post to hold some state vars: we reuse self.post.values.
        We have lost most context, and structuring of the gcode/job at this point,
        and the export_common() has potentially inserted and translated stuff.
        So, we repeat some logic from UtilsParse, like "is this a tool change?", etc.
        Line-numbers are also lost, and won't match the "inspect toolpath commands" g-code.
        We do have self.post._job
        """
        native = ""
        before_len = len(gcode.split(nl))

        # We reuse the postprocessor.values
        # So, we have to reset some values
        self.reset_values()
        Path.Log.debug(f"post.values {self.post.values}")

        for gcode_line in gcode.split(nl):

            # be nice
            if time.monotonic() - self.last_gui_update >= PreventGuiTimeout:
                if "Gui" in dir(FreeCADGui):
                    FreeCADGui.updateGui()
                self.last_gui_update = time.monotonic()

            self.post.values["line_number"] += 1  # actual line number

            # we need Path.Commands to work with
            path_command = self.to_path_command(
                gcode_line, f"expanded gcode line {self.post.values['line_number']}"
            )
            if path_command is None:
                continue

            # canonicalize to 2 digits
            if path_command.Name.startswith("("):
                pass
            else:
                # canonicalize to 2 digits
                # All the @gcode() methods can now assume 2 digits
                # We have conflated Path.Command.Name and general gcode. this allows us to expand drill, then "post process" that.
                path_command.Name = re.sub(r"^([A-Z])(\d(\.|$))", r"\g<1>0\2", path_command.Name)

            # print(f"### GCODE [{self.post.values['line_number']}] {path_command.toGCode()}")

            # And now we reproduce most of UtilsParse.parse_a_path

            self.track_by_comments(path_command)

            if self.post.arguments.gcode_comments and not path_command.Name.startswith("("):
                native += self.comment(
                    f"[{self.post.values['line_number']}] {path_command.toGCode()}"
                )

            new_location = [
                float(path_command.Parameters.get(a, self.current_location.get(a, 0.0)) or 0.0)
                for a in self.PositionAxis
            ]
            skip_modal = False

            if (
                path_command.Name in {"G00", "G01"}
                and self.post.values["MODAL"]
                and new_location == self.end_location
            ):
                skip_modal = True

            # one place to figure out our end, used by set_speed()
            if path_command.Name in self.post.values["MOTION_COMMANDS"]:
                if self.post.values["MOTION_MODE"] == "G90":
                    self.end_location = new_location
                else:
                    raise Exception("Relative G91 not supported yet")
                    # FIXME: not tested (relative mode not fully implemented):
                    # self.end_location = [
                    #    float(self.current_location[a] or 0.0) + float(path_command.Parameters.get(a, 0.0))
                    #    for a in self.current_location if a in self.PositionAxis
                    # ]
                # print(f"### end at {self.end_location}")

            if skip_modal:
                rez = ""
            else:
                # handle that gcode
                # print(f"### path_command is {path_command.__class__.__name__} {path_command}")
                rez = self.dispatch(path_command)
                # print(f"### translated: {rez.rstrip()}")

            # append to buffer
            native += rez

            if path_command.Name in self.post.values["MOTION_COMMANDS"]:
                # does the right thing for position axis, for relative
                for i, a in enumerate(self.PositionAxis):
                    self.current_location[a] = self.end_location[i]
                # all other parameters (especially F)
                for a, v in path_command.Parameters.items():
                    if a not in self.PositionAxis:
                        self.current_location[a] = v
                # print(f"### current {self.current_location}")

            self.post.values["last_command"] = path_command

        native += self.postfix()

        Path.Log.debug(f"GCode in {before_len} -> out {len(native.split(nl))}")
        return native

    def postfix(self):
        rez = ""

        if self._postfix:
            rez += nl.join(reversed(self._postfix))
            rez += nl

        return rez

    def reset_values(self):
        self.post.values.update(
            {
                "line_number": 0,
                "COMMENT_SYMBOL": "'",
            }
        )

    def track_by_comments(self, path_command):
        """Since we've lost the Path objects structure (we only have the gcode str),
        we'll track things of interest by comments here.
        """
        if path_command.Name.startswith("("):
            if m := (
                re.match(r"\(Path: ([^)]+)\)", path_command.Name)
                or re.match(r"\(Begin operation: ([^)]+)\)", path_command.Name)
                or re.match(r"\(Begin (preamble)\)", path_command.Name)
            ):
                self.post.values["Operation"] = m.group(1)

    def to_path_command(self, gcode: str, during: str):
        """Utility,
        One gcode line to a Path.Command, with some error handling
        e.g. to_path_command("G0 X50", "preamble")
        `during` is explanatory text if we get a conversion error
        """

        if gcode == "":
            return None

        pc = Path.Command()
        try:
            pc.setFromGCode(gcode)
        except ValueError as e:
            # can't tell if it is really 'Badly formatted GCode argument', so just add our gcode to the message
            message = f"During {during}, {self.location(gcode)}"
            FreeCAD.Console.PrintError(message)
            raise ValueError(message) from e
        return pc

    def dispatch(self, path_command):
        """The whole point of this class is to put the translate stuff in a modular-unit,
        and translate g-codes.
        So, each gcode maps to a function that handles it,
        so, we use @gcode decorators to collect the gcodes and functions,
        and we dispatch here.
        We return a string, which might be multi-line,
        or '' to mean nothing resulted
        """
        command = None
        if path_command.Name.startswith("("):
            # sadly, Path.Command doesn't set .Name to "comment", but rather to the comment string
            command = "comment"
        else:
            command = path_command.Name

        # Call the translate handler
        if command in self.DispatchMap:
            rez = self.DispatchMap[command](
                self, path_command
            )  # careful, doesn't do inheritance lookup
            return rez

        else:
            message = f"gcode not handled at {self.location(path_command)}"
            if (
                self.post.arguments.abort_on_unknown
                # FIXME: cf SUPPRESS_UNKNOWN
                and command not in self.post.values["ALLOW_UNKNOWN"]
            ):
                FreeCAD.Console.PrintError(message + "\n")
                raise NotImplementedError(message)
            else:
                FreeCAD.Console.PrintWarning("Skipped:  " + message + "\n")
                return self.comment(message)

    def location(self, path_command=None):
        """a message fragment of where we are, and the path_command if you want
        `path_command` can be a literal-string gcode, or usually a Path.Command
        """
        g = (
            f": {self.post.values['line_number']} {path_command if isinstance(path_command, str) else path_command.toGCode()}"
            if path_command
            else ""
        )
        return f"{self.post.values['Operation']}{g}"

    def comment(self, message, force=False):
        """if OUTPUT_COMMENTS, then generate the comment,
        `force` ignores OUTPUT_COMMENTS and always generates.
        INCLUDES the \n (nl)!
        """
        # note that comments from FreeCAD and export_common are strings with "()"
        # so, you can't tell what comments came from a Path operation object, or from Path/Post utils
        # and the parenthesis are preservered, so the final looks like '(blah blah)
        # Comments from this postprocessor are w/o (), so the final looks like 'our comment
        if self.post.values["OUTPUT_COMMENTS"] or force:
            return self.post.values["COMMENT_SYMBOL"] + message + nl
        else:
            return ""

    @gcode("comment")
    def t_comment(self, path_command):
        # leaves ()
        rez = ""

        rez += self.comment(path_command.Name)

        # fixups
        # We don't have access to the Path object, and we need/want to know where we are
        # e.g. probing. This should be fixed in the new "machine" style
        if path_command.Name.startswith("(Post Processor: "):
            rez += self.comment("  " + self.post._job.PostProcessorArgs)
        elif path_command.Name.startswith("(Cam File: "):
            rez += self.comment(f"Job: {self.post._job.Label}")
        elif m := re.match(r"\(\s*MC_RUN_COMMAND\s+(.+)\)$", path_command.Name):
            # let's leave the original as a comment (if comments are on)
            rez += m.group(1) + "\n"
        elif m := re.match(r"\(PROBEOPEN (.+)\)$", path_command.Name):
            filename = m.group(1)
            if "." not in filename:
                # default .txt (really "space delimited values")
                filename += ".txt"

            # the Probe operation

            # can't get &UserDataFolder to catenate properly anywhere...
            # so, just filename
            rez += self.comment(
                "Load the My_Variables file from Custom Cut 90 in C:\\SbParts\\Custom"
            )
            rez += "C#,90" + nl
            # if re.match(r'[^:]+:', filename):
            #    # "absolute"
            #    rez += f'OPEN "{filename}" FOR OUTPUT as #1' + nl
            # else:
            #    # "relative"
            #    rez += "GetUsrPath, &UserDataFolder" + nl
            #    rez += f'OPEN &UserDataFolder & "/{filename}" FOR OUTPUT as #1' + nl
            rez += f'OPEN "{filename}" FOR OUTPUT as #1' + nl

            rez += "&hit = 0" + nl
            # subroutines, cleanup
            # but only once per post
            if self.post.values["first_probe"]:
                self.post.values["first_probe"] = False
                self._postfix.append(
                    """GOTO SkipProbeSubRoutines
CaptureZPos:
  ' for g38.2 probe, write the data on probe-contact
  ' and set flag for didn't-fail
  ' xyzab
  WRITE #1; %(1); " "; %(2); " "; %(3); " "; %(4); " "; %(5)
  &hit = 1
  RETURN
FailedToTouch:
  ' for g38.2 probe, when
  ' failed to trigger w/in movement
  MSGBOX(Failed to touch...Exiting,16,Probe Failed)
  END
SkipProbeSubRoutines:"""
                )
        elif path_command.Name == "(PROBECLOSE)":
            rez += self.comment("Clear probe-switch-trigger")
            rez += "ON INPUT(&my_ZzeroInput, 1)" + nl
            rez += "CLOSE #1" + nl

        return rez

    @gcode("G20", "G21")  # inches, metric
    def t_units(self, path_command):
        if self.set_units:
            raise ValueError(
                "You can only set the units once, already {self.set_units['command']} at {self.set_units['at']}. You tried again at {self.location(path_command)}"
            )
        else:
            # remember where
            self.set_units = {"command": path_command.Name, "at": self.location()}

            undesired_units = {"G20": "1", "G21": "0"}[path_command.Name]  # OPPOSITE!
            rez = [
                f"IF %(25) = {undesired_units} THEN GOTO WrongUnits",
            ]

            pp_which = {"G20": "G20/--inches", "G21": "G21/--metric"}[path_command.Name]
            self._postfix.append(
                nl.join(
                    [
                        "GOTO AfterWrongUnits",
                        "WrongUnits:",
                        '  if %(25) = 0 THEN &shopbot_which="inches"',
                        '  if %(25) = 1 THEN &shopbot_which="mm"',
                        # NB: no commas in strings!
                        f'    MSGBOX("Post-processor wants {pp_which} but ShopBot is " & &shopbot_which & ". Change Units in ShopBot and try again.",0,"Change Units")',
                        "    ENDALL",
                        "AfterWrongUnits:",
                    ]
                )
            )

            return nl.join(rez) + nl

    @gcode("G90")  # no relative (G91) yet, have to fix modal handling for relative
    def t_absolute_mode(self, path_command):
        self.post.values["MOTION_MODE"] = path_command.Name
        return {"G90": "SA", "G91": "SR"}[path_command.Name] + nl

    @gcode("M06")
    def t_toolchange(self, path_command):
        tool_number = int(path_command.Parameters["T"])

        # check for tool actually existing
        tool_controller = next(
            (x for x in self.post._job.Tools.Group if x.ToolNumber == tool_number), None
        )
        if not tool_controller:
            # HACK: at least till 1.1, nothing enforces tool-numbers in the job to be unique
            #   and "Tn" doesn't have to match a ToolNumber
            #   we'll do a compatibility hack ONLY if all tools == 1
            if (
                all(x.ToolNumber == 1 for x in self.post._job.Tools.Group)
                and len(self.post._job.Tools.Group) >= tool_number
            ):
                tool_controller = self.post._job.Tools.Group[tool_number - 1]
                FreeCAD.Console.PrintWarning(
                    f"Job <{self.post._job.Label}> doesn't have unique tool-numbers? at {self.location(path_command)}"
                )
            else:
                raise ValueError(
                    f"Toolchange with non-existent tool_number {tool_number} at {self.location(path_command)}. Do tools have unique tool-numbers?"
                )

        tool_name = f"{tool_controller.Label}, {tool_controller.Tool.Label}"  # not sure if we want both .Label's, just trying to help the operator
        safe_tool_name = re.sub(r"[^A-Za-z0-9/_ .-]", "", tool_name)

        rez = []

        if not self.post.values["OUTPUT_TOOL_CHANGE"]:
            rez.append(
                self.comment(
                    f"First change tool, should already be #{tool_number}: {safe_tool_name}",
                    force=True,
                ).rstrip()
            )

        rez += [
            f"&Tool={tool_number}",
            f'&ToolName="{safe_tool_name}"',
        ]

        if self.post.values["OUTPUT_TOOL_CHANGE"]:
            # automatic no prompt, or manual prompt (depends on correct shopbot setup)
            rez.append("C9")

        else:
            if self.first_tool:
                self.first_tool = False
            else:
                raise NotImplementedError(
                    f"2nd tool can't be done, #{tool_number}, no way to change-tool when --no-tool-changer at {self.location(path_command)}. Try 'Order by Tool' or 'Order by Operation' in job's 'Output' tab."
                )

        rez.append(self.set_initial_speeds(tool_controller, path_command).rstrip())
        rez = nl.join((x for x in rez if x != "")) + nl

        return rez

    @gcode("G00", "G01")
    def t_move(self, path_command):
        """Oh boy.
        opensbp specifies the xy speed, and Z speed separately for a motion.
        e.g. a "VS,sxy,sz" then the move like "M3,x,y,z".
        But, gcode has a F which the speed of the vector
        (for rapid, it's whatever-the-machine-setting-is).
        FreeCAD has horizontal speed (xy), and vertical speed (z),
        which it uses to calculate F,
        (we already set rapid at set_initial_speed() time for each tool-change).
        Finally, we have to take the delta(x,y,z) vector and project the F speed onto xy, and z
        to generate the MS command before each move command.
        The Mx or Jx just uses the axis distances.
        FIXME? If the first motion is not G0, and axis are 'Z' + X|Y,
            (e.g. G1 X10 Y20 Z30)
            then we don't know how to calculate the MS speed,
            because we need to know the last-position to split F across Z & X|Y,
            and there isn't one (that we know about), though we init'd to 0,0,0.
            We assume a G1 happens before any other motion, to establish a position.
            We could abort on this...
        """
        rez = ""

        # Optimize the command, specifying 1..5 axis values
        axis = [path_command.Parameters.get(a, None) for a in self.PositionAxis]
        last_not_none = 0
        # XYZABC, but reversed
        for i in reversed(range(0, len(axis))):
            if axis[i] is not None:
                last_not_none = i
                break
        axis = axis[: last_not_none + 1]

        if feed_rate := path_command.Parameters.get("F", None):
            if path_command.Name == "G00":
                if self.post.arguments.abort_on_unknown:
                    raise ValueError(
                        f"Rapid moves (G0) can't have an F at {self.location(path_command)}"
                    )

        _, speed_command = self.set_speed(path_command)
        rez += speed_command

        native_command = "J" if path_command.Name == "G00" else "M"
        # print(f"  ### is a '{native_command}'")

        # nb, we don't have to do anything for --axis-modal, handled by common stuff earlier!

        axis_ct = len(axis)
        if axis_ct == 1:
            native_command += "X"
        else:
            native_command += str(axis_ct)

        formatted_axis = (
            (format(a, f".{self.post.values['FEED_PRECISION']}f") if a is not None else "")
            for a in axis
        )
        rez += f"{native_command},{','.join(formatted_axis)}" + nl

        return rez

    @gcode("M03")  # clockwise only. do the spindle-controlers do CCW?
    def t_spindle_speed(self, path_command):
        native = ""

        if "S" in path_command.Parameters:
            native += f"TR,{int(path_command.Parameters['S'])}\n"  # rpm units

        # macro will do the dialog-box if you don't have a controlled spindle
        native += "C6\n"

        if self.post.values["SPINDLE_WAIT"] > 0:
            native += f"PAUSE {int(self.post.values['SPINDLE_WAIT'])}\n"

        return native

    def format_value(self, value, precision_type="FEED_PRECISION"):
        """format for the precision (e.g. AXIS_PRECISION)
        notably dealing with slightly-less-than-zero == "0" not "-0"
        """
        # format rounds, so duplicate that effect
        if abs(value) < 0.5 * 10 ** (-self.post.values[precision_type]):
            value = 0.0
        return format(value, f".{self.post.values[precision_type]}f")

    @gcode("G02", "G03")
    def t_arc(self, path_command):
        # only center-format: IJ
        # only absolute mode
        # only xy plane
        # only P1 (or no P)
        # cases:
        #   current-position is start
        #   Z causes helical-arc
        #   Pn causes arc-as-defined + (n-1) whole circles: not handled
        #   XY is the end-position for a segment
        #   none of XY means whole circle
        #   IJ is the location of the arc-center: an offset. at least one of is required
        #   F is required

        # we would have to generate multiple CG's for repetitions (P)
        handled_parameters = "XYZIJFPK"  # notably, not R

        not_handled = []
        not_handled = [a for a in path_command.Parameters if a not in handled_parameters]
        # we handle K=0.0 by ignoring it (xy-plane), other K's we don't handle
        if "K" in path_command.Parameters and path_command.Parameters["K"] != 0.0:
            not_handled.append("K")
        if "P" in path_command.Parameters and path_command.Parameters["P"] != 1:
            not_handled.append("P")
        if not_handled:
            message = (
                f"We can't do parameters {not_handled} for an arc in {self.location(path_command)}"
            )
            FreeCAD.Console.PrintError(message)
            if self.post.arguments.abort_on_unknown:
                raise ValueError(message)
            else:
                return ""

        if self.post.values["MOTION_MODE"] != "G90":
            opname = self.post.values["Operation"].Label if self.post.values["Operation"] else ""
            message = f"We can't do relative mode for arcs in [{self.post.values['line_number']}] {opname} {path_command.toGCode()}"
            FreeCAD.Console.PrintError(message)
            if self.post.arguments.abort_on_unknown:
                raise NotImplementedError(message)
            else:
                return ""

        if path_command.Name == "G02":  # CW
            dirstring = "1"
        else:  # G3 means CCW
            dirstring = "-1"
        txt = ""

        dz, speed_command = self.set_speed(path_command)
        txt += speed_command

        txt += "CG,"
        txt += ","  # no diameter

        # end
        # Omitting XY has special meaning to ShopBot, it is not the same as modal-axis
        # The PostProcess code will drop the XYZ axis on --axis-modal, but we need it:
        x = path_command.Parameters.get("X", self.current_location["X"])
        y = path_command.Parameters.get("Y", self.current_location["Y"])

        txt += self.format_value(x) + ","
        txt += self.format_value(y) + ","

        # Center is at offset:
        txt += (
            self.format_value(
                path_command.Parameters["I"] if "I" in path_command.Parameters.keys() else 0.0
            )
            + ","
        )
        txt += (
            self.format_value(
                path_command.Parameters["J"] if "J" in path_command.Parameters.keys() else 0.0
            )
            + ","
        )
        txt += "T" + ","  # move on diameter
        txt += dirstring + ","

        if "Z" not in path_command.Parameters:
            dz = 0

        # Z causes a helical, "causes the defined plunge to be made gradually as the cutter is circling down"
        # Note, dz is actual distance vector, but ShopBot uses -dz to mean "plunge" relative
        txt += self.format_value(-dz) + ","

        txt += ","  # repetitions
        txt += ","  # proportion-x
        txt += ","  # proportion-y

        if dz != 0.0:
            # helical cases
            # we don't do "bottom pass" (4) because FreeCAD seems to do that and it's not a g-code thing anyway
            if "X" not in path_command.Parameters and "Y" not in path_command.Parameters:
                # circle
                feature = 3  # spiral
            else:
                feature = 3  # spiral
        else:
            feature = 0

        txt += f"{feature},"
        txt += "1,"  # continue the CG plunging (don't pull up)
        txt += "0"  # no move before plunge

        # actual Z, opensbp plunge is a delta, note the actual Z as a comment
        z = path_command.Parameters.get("Z", self.current_location["Z"])
        txt += " ' Z" + self.format_value(z)
        txt += "\n"
        return txt

    @gcode("M00", "M01")
    def t_prompt(self, command):
        # Prompt with "Continue?" and pause, wait for user-interaction
        # If a comment precedes M00, that is used as the prompt (to emulate opensbp behavior)

        txt = ""
        if not self.post.values["last_command"].Name.startswith("("):
            # default prompt
            where = []
            if self.post._job:
                where.append(f"<{self.post._job.Label}>")
            if self.post.values["Operation"]:
                where.append(f"<{self.post.values['Operation']}>")
            txt += self.comment(f"Continue {'.'.join(where)}?", force=True)
        elif not self.post.values["OUTPUT_COMMENTS"]:
            # Force inclusion of that preceding comment as a prompt
            txt += self.comment(self.post.values["last_command"].Name, force=True)
        txt += "PAUSE\n"
        return txt

    @gcode("M05")
    def t_stop_spindle(self, command):
        return "C7\n"

    @gcode("M02", "M30")
    def t_stop(self, command):
        return "END\n"

    @gcode("M08")
    def t_coolant_on(self, command):
        return "SO,3,1\n"

    @gcode("M09")
    def t_cooland_off(self, command):
        return "SO,3,0\n"

    @gcode("G38.2")
    def t_probe(self, command):

        speed = command.Parameters.get("F", None)
        if speed is not None:
            speed = float(speed)
        if speed == 0.0:
            FreeCAD.Console.PrintWarning(
                f"G38.2 with an F0.0, set Tool speeds? at {self.location(command)}\n"
            )
            return ""
        if speed is None:
            speed = self.current_location["ms"][1]  # out of [ xy, z ]

        axis = " ".join(
            [f"{a}{command.Parameters[a]}" for a in self.PositionAxis if a in command.Parameters]
        )

        # PROBEOPEN sets up the contact-detect, so G38.2 are just moves
        rez = ""

        rez += "&hit = 0" + nl  # for did-we-hit OR fail

        # for probing, we have to setup the on-input for every move
        rez += "ON INPUT(&my_ZzeroInput, 1) GOSUB CaptureZPos" + nl

        g = f"G01 F{speed} {axis}"
        rez += self.t_move(Path.Command(g))

        # and check for fail to contact
        rez += "IF &hit = 0 THEN GOTO FailedToTouch\n"

        return rez

    def set_speed(self, path_command):
        # For non-rapid, F applies to the vector of all the axis
        # For rapid, full speed on the axis from the toolchange settings
        #   (so no output here)
        # Projects F into XY plane, and Z plane for shopbot
        # Always uses "VS" command to not "punctuate" the stack (ms/js will cause ramp up/down in speed)
        # Elides VS values if not change or 0.0 (shopbot doesn't like 0)
        # Elides trailing ,
        # Elides whole VS if no values

        if path_command.Name == "G00":
            # Actually, we just use the full speed on xy and z axis
            # which was initialized at toolchange time
            # the args to vs would be: VS,xy,z,a,b,xy_job,z_job,a_job,b_jog
            return (0, "")

        # we output VS, but this lets us track MS vs JS
        which_speed = "MS"

        last_position = [float(self.current_location[a] or 0) for a in self.PositionAxis]

        def fmt_diff(l):
            return [(f"{p:9.3f}" if p is not None else f"{str(p):9s}") for p in l]

        # Linear move
        if path_command.Name == "G01":
            d_axis = list(map(operator.sub, self.end_location, last_position))

            squared_d_axis = [v**2 for v in d_axis]
            distance = math.sqrt(sum(squared_d_axis))
            z_distance = abs(d_axis[2])
            xy_distance = math.sqrt(sum(squared_d_axis[:2]))
            axis = [a for a in self.PositionAxis if a in path_command.Parameters]

        # Arcs
        elif path_command.Name in {"G02", "G03"}:

            def arc_length_3d(center, start, end, clockwise):
                """center, start, end: (x, y, z) tuples
                clockwise: True for G2, False for G3
                Returns the true 3D arc length.
                """

                cx, cy, cz = center
                sx, sy, sz = start
                ex, ey, ez = end

                # ---- linear Z interpolation ----
                dz = ez - sz

                # ---- XY arc angle ----
                r = math.hypot(sx - cx, sy - cy)

                a0 = math.atan2(sy - cy, sx - cx)
                a1 = math.atan2(ey - cy, ex - cx)

                dtheta = a1 - a0

                if dtheta == 0:
                    dtheta = 2 * math.pi
                elif clockwise:
                    if dtheta > 0:
                        dtheta -= 2 * math.pi
                else:
                    if dtheta < 0:
                        dtheta += 2 * math.pi

                arc_xy = abs(r * dtheta)

                # ---- true helical arc length ----
                return (arc_xy, math.hypot(arc_xy, dz))

            start_position = [float(self.current_location[a] or 0) for a in "XYZ"]
            center_offset = [
                float(path_command.Parameters.get(a, None) or 0) for a in "IJK"
            ]  # k always 0
            end_position = [
                float(path_command.Parameters.get(k, start_position[i]))
                for i, k in enumerate("XYZ")
            ]
            z_distance = start_position[2] - end_position[2]

            # If the XY is omitted, it means a whole circle, and arc-length-3d will give that
            xy_distance, distance = arc_length_3d(
                map(operator.add, start_position, center_offset),  # center
                start_position,
                end_position,
                path_command.Name == "G02",  # clockwise?
            )

            axis = list("XYZ")

        distances_for_speed = [abs(xy_distance), abs(z_distance)]  # abs(xy) shouldn't be necessary

        if path_command.Name == "G00":
            # see above, we don't get to here on G0
            # Rapid speed is the full speed on xy, and z
            # no projecting on to the vector
            # FIXME: AB not handled yet
            speeds = self.current_location["js"]

        # feed motions
        else:
            # FIXME: AB speeds not handled yet

            f = path_command.Parameters.get("F", None)
            if f is None:
                f = self.current_location["F"]
                # print(f"  ### no f, last= '{f}'")

            if f is None:
                # No F and no previous, which is not good. default to machine's feed speeds.
                f = None
                if self.first_no_F:
                    FreeCAD.Console.PrintWarning(
                        f"No F, and no previous F speed at {self.location(path_command)}. Using tool's feed speeds.\n"
                    )
                    self.first_no_F = False
                speeds = ["", ""]

            else:
                # have a F/previous: use it
                f = float(f)

                # if only in xy, or only in z, then speed=F
                if xy_distance != 0.0 and z_distance == 0.0:
                    speeds = [f, 0.0]
                elif xy_distance == 0.0 and z_distance != 0.0:
                    speeds = [0.0, f]

                else:
                    # FIXME: AB not handled yet
                    speeds = [
                        ((f * d / distance) if distance != 0 else 0) for d in distances_for_speed
                    ]
                    # print(f"  ### speed w/xyz {speeds}")
            # print(f"  ### speeds  {speeds}")

        min_speed = self.post.values["MIN_SPEED"]

        def gtmin(s):
            if s == "":
                return s
            elif abs(s) >= min_speed:
                return s
            elif s == 0.0:
                return ""
            elif abs(s) < min_speed:
                return min_speed * (-1 if s < 0 else 1)
            # that's all the cases

        speeds = [gtmin(s) for s in speeds]
        speeds = [
            (format(s, f'.{self.post.values["SPEED_PRECISION"]}f') if s != "" else "")
            for s in speeds
        ]

        if which_speed == "MS":
            # axis-modal against actual output (formatted to SPEED_PRECISION)
            # and save as current

            non_elided = copy(speeds)

            if True:  # speed-modal always true
                # compare to previous 'ms' speeds, so we can skip 'MS' if nothing changes
                for i, new_speed in enumerate(speeds):
                    old_speed = self.current_location[which_speed.lower()][i]
                    if old_speed == new_speed:
                        speeds[i] = ""

            # save it for next time, for --speed-modal
            self.current_location["ms"] = non_elided

        # cleans up trailing , when trailing speeds elided
        if which_speed == "MS":
            cmd = f"VS,{','.join(speeds)}".rstrip(",")
        elif which_speed == "JS":
            # again, shouldn't get here in this version of the code
            cmd = f"VS,,,,,{','.join(speeds)}".rstrip(",")

        # If there is no speed to set (e.g. the move ends up as delta-0), no MS needed
        if cmd == "VS":
            cmd = ""
        else:
            cmd += nl
        return (z_distance, cmd)

    def set_initial_speeds(self, tool_controller, path_command):
        # need to ensure initial values for speeds
        # rapid-speed is never emitted by gcode, but we need to set it!
        # and we just set the initial speed for "feed" too

        Path.Log.debug(f"Setspeeds {tool_controller.Label}")
        native = ""

        native += self.comment(f"set speeds: {tool_controller.Label}")
        speeds = {
            "ms": [],  # xy,z
            "js": [],  # xy,z
        }

        def append_speed(which_speed, which_speed_key):
            with_units = getattr(tool_controller, which_speed)
            speed = float(with_units.getValueAs(self.post.values["UNIT_SPEED_FORMAT"]))

            if abs(speed) >= 0.5 * 10 ** (-self.post.values["SPEED_PRECISION"]):  # i.e. not zero

                formatted = format(speed, f'.{self.post.values["SPEED_PRECISION"]}f')

                speeds[which_speed_key].append(formatted)
                return None
            else:
                FreeCAD.Console.PrintWarning(
                    f"ToolController <{self.post._job.Label}>.<{tool_controller.Label}> did not set {which_speed} speed, set the HorizFeed and VertFeed. ( for {self.location(path_command)} )\n"
                )
                speeds[which_speed_key].append("")
                if which_speed.endswith("Rapid"):
                    warn_rapid.append(which_speed)
                return self.comment(f"no {which_speed}", force=True)

        warn_rapid = []  # empty is no warning

        # tool's speeds -> speeds[ 'ms' & 'js' ]
        for which_speed_key, which_speed_properties in {
            "ms": ["HorizFeed", "VertFeed"],
            "js": ["HorizRapid", "VertRapid"],
        }.items():
            for which_speed in which_speed_properties:  # by property for warning messages
                comment = append_speed(which_speed, which_speed_key)
                if comment is not None:
                    native += comment

            # for --speed-modal
            self.current_location[which_speed_key] = speeds[which_speed_key]

            # add to command-stream
            command_prefix = which_speed_key.upper()
            command = (f"{command_prefix}," + ",".join(speeds[which_speed_key])).rstrip(",")
            if command != command_prefix:  # has actual speeds
                native += command + "\n"

        # fixme: where to get A&B values?
        # speeds['ms'].append('') # a-move-speed
        # speeds['ms'].append('') # b-move-speed

        if warn_rapid:
            if not self.post.arguments.native_rapid_fallback:
                raise ValueError(
                    f"ToolController <{self.post._job.Label}>.<{tool_controller.Label}> did not set xy&z rapid speeds, and you specified --no-native-rapid-fallback. Set the rapid speeds. {self.location(path_command)}"
                )
            else:
                FreeCAD.Console.PrintWarning(
                    f'Using machine\'s rapid ("jog") for {" and ".join(warn_rapid)}, for ToolController <{self.post._job.Label}>.<{tool_controller.Label}>\n'
                )

        Path.Log.debug(f"setspeeds {speeds}")

        return native


gcode_insertmap()  # fixup DispatchMap



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from lazy_loader.lazy_loader import LazyLoader
import Path
import Path.Op.Base as PathOp
import Path.Op.Util as PathOpUtil
import PathScripts.PathUtils as PathUtils

# import copy

__doc__ = "Base class for all ops in the engrave family."

# lazily loaded modules
DraftGeomUtils = LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")
Part = LazyLoader("Part", globals(), "Part")

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class ObjectOp(PathOp.ObjectOp):
    """Proxy base class for engrave operations."""

    def getZValues(self, obj):
        zValues = []
        if obj.StepDown.Value != 0:
            z = obj.StartDepth.Value - obj.StepDown.Value
            stepdown = obj.StepDown.Value
            if stepdown < 0:
                stepdown = -stepdown

            while z > obj.FinalDepth.Value:
                zValues.append(z)
                z -= stepdown

        zValues.append(obj.FinalDepth.Value)
        return zValues

    def buildpathocc(self, obj, wires, zValues, relZ=False, forward=True, start_idx=0):
        """buildpathocc(obj, wires, zValues, relZ=False) ... internal helper function to generate engraving commands."""
        Path.Log.track(obj.Label, len(wires), zValues)

        # sort wires, adapted from Area.py
        if len(wires) > 1:
            locations = []
            for w in wires:
                locations.append({"x": w.BoundBox.Center.x, "y": w.BoundBox.Center.y, "wire": w})

            locations = PathUtils.sort_locations(locations, ["x", "y"])
            wires = [j["wire"] for j in locations]

        decomposewires = []
        for wire in wires:
            decomposewires.extend(PathOpUtil.makeWires(wire.Edges))

        wires = decomposewires
        for wire in wires:
            # offset = wire

            # reorder the wire
            if hasattr(obj, "StartVertex"):
                start_idx = obj.StartVertex
            edges = wire.Edges

            # edges = copy.copy(PathOpUtil.orientWire(offset, forward).Edges)
            # Path.Log.track("wire: {} offset: {}".format(len(wire.Edges), len(edges)))
            # edges = Part.sortEdges(edges)[0]
            # Path.Log.track("edges: {}".format(len(edges)))

            last = None

            for z in zValues:
                Path.Log.debug(z)
                if last and wire.isClosed():
                    # Add step down to next Z for closed profile
                    self.appendCommand(
                        Path.Command("G1", {"X": last.x, "Y": last.y, "Z": last.z}),
                        z,
                        relZ,
                        self.vertFeed,
                    )

                first = True
                if start_idx > len(edges) - 1:
                    start_idx = len(edges) - 1

                edges = edges[start_idx:] + edges[:start_idx]
                for edge in edges:
                    Path.Log.debug(
                        "points: {} -> {}".format(edge.Vertexes[0].Point, edge.Vertexes[-1].Point)
                    )
                    Path.Log.debug(
                        "valueat {} -> {}".format(
                            edge.valueAt(edge.FirstParameter),
                            edge.valueAt(edge.LastParameter),
                        )
                    )
                    if first and (not last or not wire.isClosed()):
                        Path.Log.debug("processing first edge entry")
                        # Add moves to first point of wire
                        last = edge.Vertexes[0].Point

                        self.commandlist.append(
                            Path.Command(
                                "G0",
                                {"Z": obj.ClearanceHeight.Value, "F": self.vertRapid},
                            )
                        )
                        self.commandlist.append(
                            Path.Command("G0", {"X": last.x, "Y": last.y, "F": self.horizRapid})
                        )
                        self.commandlist.append(
                            Path.Command("G0", {"Z": obj.SafeHeight.Value, "F": self.vertRapid})
                        )
                        self.appendCommand(
                            Path.Command("G1", {"X": last.x, "Y": last.y, "Z": last.z}),
                            z,
                            relZ,
                            self.vertFeed,
                        )
                    first = False

                    if Path.Geom.pointsCoincide(last, edge.valueAt(edge.FirstParameter)):
                        # if Path.Geom.pointsCoincide(last, edge.Vertexes[0].Point):
                        # Edge not reversed
                        for cmd in Path.Geom.cmdsForEdge(edge):
                            # Add gcode for edge
                            self.appendCommand(cmd, z, relZ, self.horizFeed)
                        last = edge.Vertexes[-1].Point
                    else:
                        # Edge reversed
                        for cmd in Path.Geom.cmdsForEdge(edge, True):
                            # Add gcode for reversed edge
                            self.appendCommand(cmd, z, relZ, self.horizFeed)
                        last = edge.Vertexes[0].Point

            self.commandlist.append(
                Path.Command("G0", {"Z": obj.ClearanceHeight.Value, "F": self.vertRapid})
            )

    def appendCommand(self, cmd, z, relZ, feed):
        params = cmd.Parameters
        if relZ:
            z = params["Z"] - z
        params.update({"Z": z, "F": feed})
        self.commandlist.append(Path.Command(cmd.Name, params))

    def opSetDefaultValues(self, obj, job):
        """opSetDefaultValues(obj) ... set depths for engraving"""
        if PathOp.FeatureDepths & self.opFeatures(obj):
            if job and len(job.Model.Group) > 0:
                bb = job.Proxy.modelBoundBox(job)
                obj.OpStartDepth = bb.ZMax
                obj.OpFinalDepth = bb.ZMax - max(obj.StepDown.Value, 0.1)
            else:
                obj.OpFinalDepth = -0.1
